import copy
import logging

from mpf.platforms.opp.opp import OppHardwarePlatform
from unittest.mock import MagicMock

import time

from mpf.platforms.opp import opp
from mpf.platforms.opp.opp_rs232_intf import OppRs232Intf
from mpf.tests.MpfTestCase import MpfTestCase
from mpf.tests.loop import MockSerial

COMMAND_LENGTH = {
    0x00: 7,
    0x02: 7,
    0x07: 7,
    0x08: 7,
    0x0d: 7,
    0x13: 8,
    0x14: 7,
    0x17: 5,
    0x19: 11,
}

class MockOppSocket(MockSerial):

    def read(self, length):
        del length
        if not self.queue:
            return b""
        msg = b""
        while self.queue:
            msg += self.queue.pop(0)
        return msg

    def read_ready(self):
        return bool(self.queue)

    def write_ready(self):
        return True

    def write(self, msg):
        """Handle messages in fake OPP."""
        #print("Serial received: " + "".join("\\x%02x" % b for b in msg) + " len: " + str(len(msg)))
        total_msg_len = len(msg)
        if self.crashed:
            return

        while msg:
            # special case: EOM and inventory map
            if msg[0] in (0xff, 0xf0):
                self._handle_msg(msg[0:1])
                msg = msg[1:]
                continue

            if len(msg) < 2:
                raise AssertionError("Message too short. " + "".join("\\x%02x" % b for b in msg))

            command = msg[1]
            if command == 0x40:
                # special case of variable length message
                if len(msg) < 6:
                    raise AssertionError("Fade too short. " + "".join("\\x%02x" % b for b in msg))
                command_len = 9 + msg[5]
            else:
                if command not in COMMAND_LENGTH:
                    raise AssertionError("Unknown command. " + "".join("\\x%02x" % b for b in msg))
                command_len = COMMAND_LENGTH[command]

            if len(msg) < command_len:
                raise AssertionError("Command length ({}) does not match message length ({}). {}".format(
                    command_len, len(msg), "".join("\\x%02x" % b for b in msg)
                ))

            self._handle_msg(msg[0:command_len])
            msg = msg[command_len:]

        return total_msg_len

    def _handle_msg(self, msg):
        # print("Handling: " + "".join("\\x%02x" % b for b in msg) + " len: " + str(len(msg)))
        if msg in self.permanent_commands:
            self.queue.append(self.permanent_commands[msg])
            return len(msg)

        if msg not in self.expected_commands:
            self.crashed = True
            remaining_expected_commands = dict(self.expected_commands)
            self.expected_commands = {"crashed": True}
            raise AssertionError("Unexpected command: " + "".join("\\x%02x" % b for b in msg) +
                                 " len: " + str(len(msg)) + " Remaining expected commands: " +
                                 str(remaining_expected_commands) + " on " + self.name)

        if self.expected_commands[msg] is not False:
            self.queue.append(self.expected_commands[msg])

        del self.expected_commands[msg]

    def __init__(self, name):
        super().__init__()
        self.name = name
        self.expected_commands = {}
        self.queue = []
        self.permanent_commands = {}
        self.crashed = False


class OPPCommon(MpfTestCase):

    def __init__(self, methodName):
        super().__init__(methodName)
        self.expected_duration = 2
        self.serialMock = None

    def get_machine_path(self):
        return 'tests/machine_files/opp/'

    def _crc_message(self, msg, term=False):
        crc_msg = msg + OppRs232Intf.calc_crc8_part_msg(msg, 0, len(msg))
        if term:
            crc_msg += b'\xff'
        return crc_msg

    def _mock_loop(self):
        self.clock.mock_serial("com1", self.serialMock)

    def tearDown(self):
        self.assertFalse(self.serialMock.crashed)
        super().tearDown()

    def get_platform(self):
        return False

    def _wait_for_processing(self):
        start = time.time()
        while self.serialMock.expected_commands and not self.serialMock.crashed and time.time() < start + 10:
            self.advance_time_and_run(.01)

        self.assertFalse(self.serialMock.crashed)


class TestOPPStm32(MpfTestCase):

    def __init__(self, methodName):
        super().__init__(methodName)
        self.expected_duration = 2
        self.serialMocks = {}

    def get_machine_path(self):
        return 'tests/machine_files/opp/'

    def _crc_message(self, msg, term=False):
        crc_msg = msg + OppRs232Intf.calc_crc8_part_msg(msg, 0, len(msg))
        if term:
            crc_msg += b'\xff'
        return crc_msg

    def _mock_loop(self):
        self.clock.mock_serial("com1", self.serialMocks["com1"])
        self.clock.mock_serial("com2", self.serialMocks["com2"])

    def tearDown(self):
        for port, mock in self.serialMocks.items():
            self.assertFalse(mock.crashed, "Mock {} crashed".format(port))
        super().tearDown()

    def get_platform(self):
        return False

    def _wait_for_processing(self):
        start = time.time()
        while sum([len(mock.expected_commands) for mock in self.serialMocks.values()]) and \
                not sum([mock.crashed for mock in self.serialMocks.values()]) and time.time() < start + 10:
            self.advance_time_and_run(.01)
        self.assertFalse(self.serialMocks["com1"].expected_commands)
        self.assertFalse(self.serialMocks["com2"].expected_commands)

    def get_config_file(self):
        return 'config_stm32.yaml'

    def setUp(self):
        self.expected_duration = 1.5
        opp.serial_imported = True
        opp.serial = MagicMock()
        self.serialMocks["com1"] = MockOppSocket("com1")
        self.serialMocks["com2"] = MockOppSocket("com2")
        board1_config = b'\x20\x0d\x01\x02\x03\x08'      # wing1: solenoids, wing2: inputs, wing3: lamps, wing4: neo_sol
        board2_config = b'\x20\x0d\x0b\x0c\x03\x03'      # wing1: lamps, wing2: lamps, wing3: lamps, wing4: lamps
        board1_version = b'\x20\x02\x02\x01\x00\x00'     # 2.1.0.0
        board2_version = b'\x20\x02\x02\x01\x00\x00'     # 2.1.0.0
        inputs1_message = b"\x20\x08\x00\xff\x00\x0c"    # inputs 0+1 off, 2+3 on, 8 on
        inputs2_message = b"\x20\x08\x00\x00\x00\x00"

        self.serialMocks["com1"].expected_commands = {
            b'\xf0': b'\xf0\x20',     # boards 20 installed
            self._crc_message(b'\x20\x0d\x00\x00\x00\x00'): self._crc_message(board1_config), # get config
            self._crc_message(b'\x20\x02\x00\x00\x00\x00'): self._crc_message(board1_version),   # get version
            self._crc_message(b'\x20\x00\x00\x00\x00\x00'): self._crc_message(b'\x20\x00\x01\x23\x45\x67'),
            self._crc_message(b'\x20\x40\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00\x00\x00'): False
        }
        self.serialMocks["com1"].permanent_commands = {
            b'\xff': b'\xff',
            self._crc_message(b'\x20\x08\x00\x00\x00\x00'): self._crc_message(inputs1_message),  # read inputs
        }

        self.serialMocks["com2"].expected_commands = {
            b'\xf0': b'\xf0\x20',     # boards 20 installed
            self._crc_message(b'\x20\x0d\x00\x00\x00\x00'): self._crc_message(board2_config), # get config
            self._crc_message(b'\x20\x02\x00\x00\x00\x00'): self._crc_message(board2_version),   # get version
            self._crc_message(b'\x20\x00\x00\x00\x00\x00'): self._crc_message(b'\x20\x00\x00\x00\x00\x02'),
            self._crc_message(b'\x20\x40\x10\x10\x00\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'): False,
            self._crc_message(b'\x20\x40\x10\x1f\x00\x01\x00\x00\x00'): False,
            self._crc_message(b'\x20\x40\x20\x00\x00\x02\x00\x00\x00\x00'): False,
            self._crc_message(b'\x20\x40\x20\x3f\x00\x01\x00\x00\x00'): False,
        }
        self.serialMocks["com2"].permanent_commands = {
            b'\xff': b'\xff',
            self._crc_message(b'\x20\x08\x00\x00\x00\x00'): self._crc_message(inputs2_message),  # read inputs
        }
        super().setUp()

        assert isinstance(self.machine.default_platform, OppHardwarePlatform)

        self._wait_for_processing()
        self.assertEqual(0x02010000, self.machine.default_platform.min_version["19088743"])
        self.assertEqual(0x02010000, self.machine.default_platform.min_version["2"])

        self.maxDiff = 100000

        # test hardware scan
        info_str = """Connected CPUs:
 - Port: com1 at 115200 baud. Chain Serial: 19088743
 -> Board: 0x20 Firmware: 0x2010000
 - Port: com2 at 115200 baud. Chain Serial: 2
 -> Board: 0x20 Firmware: 0x2010000

Incand cards:
 - Chain: 19088743 Board: 0x20 Card: 0 Numbers: [16, 17, 18, 19, 20, 21, 22, 23]
 - Chain: 2 Board: 0x20 Card: 0 Numbers: [16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31]

Input cards:
 - Chain: 19088743 Board: 0x20 Card: 0 Numbers: [0, 1, 2, 3, 8, 9, 10, 11, 12, 13, 14, 15, 25, 26, 27]

Solenoid cards:
 - Chain: 19088743 Board: 0x20 Card: 0 Numbers: [0, 1, 2, 3, 12, 13, 14, 15]

LEDs:
 - Chain: 19088743 Board: 0x20 Card: 0
"""
        self.assertEqual(info_str, self.machine.default_platform.get_info_string())

    def testOpp(self):
        # assert that the watchdog does not trigger on incand only boards
        with self.assertLogs('OPP', level='WARNING') as cm:
            self.advance_time_and_run(1)
            self.assertFalse(cm.output)
            # log something to prevent the test from breaking
            logging.getLogger("OPP").warning("DEBUG")

        # set color of neo pixel
        self.serialMocks["com1"].expected_commands[
            self._crc_message(b'\x20\x40\x00\x00\x00\x06\x00\x64\xff\x00\x00\x00\x00\xff', False)] = False
        self.machine.lights["l_neo_0"].color("red", fade_ms=100)
        self.machine.lights["l_neo_1"].color("blue", fade_ms=100)
        self.advance_time_and_run(.01)
        self._wait_for_processing()

        self.advance_time_and_run(.15)

        self.serialMocks["com1"].expected_commands[
            self._crc_message(b'\x20\x40\x00\x00\x00\x06\x00\x64\x00\x00\xff\xff\x00\x00', False)] = False
        self.machine.lights["l_neo_0"].color("blue", fade_ms=100)
        self.machine.lights["l_neo_1"].color("red", fade_ms=100)
        self.advance_time_and_run(.01)
        self._wait_for_processing()

        self.advance_time_and_run(.15)

        self.machine.lights["l_neo_0"].color("blue", fade_ms=100)
        self.machine.lights["l_neo_1"].color("red", fade_ms=100)
        self.advance_time_and_run(.01)
        self._wait_for_processing()

        self.serialMocks["com2"].expected_commands[
            self._crc_message(b'\x20\x40\x10\x13\x00\x02\x00\x64\x99\xe5', False)] = False

        self.machine.lights["l2-3"].color("white%60", fade_ms=100)
        self.machine.lights["l2-4"].color("white%90", fade_ms=100)
        self.advance_time_and_run(.02)
        self._wait_for_processing()

        self.serialMocks["com2"].expected_commands[
            self._crc_message(b'\x20\x40\x20\x00\x00\x02\x00\x64\x99\xe5', False)] = False

        self.machine.lights["m0-0"].color("white%60", fade_ms=100)
        self.machine.lights["m0-1"].color("white%90", fade_ms=100)
        self.advance_time_and_run(.02)
        self._wait_for_processing()


class TestOPPFirmware2(OPPCommon, MpfTestCase):

    def get_config_file(self):
        return 'config2.yaml'

    def setUp(self):
        self.expected_duration = 1.5
        opp.serial_imported = True
        opp.serial = MagicMock()
        self.serialMock = MockOppSocket("com1")
        board1_config = b'\x20\x0d\x01\x02\x03\x03'     # wing1: solenoids, wing2: inputs, wing3: lamps, wing4: lamps
        board2_config = b'\x21\x0d\x06\x02\x02\x01'     # wing1: neo, wing2: inputs, wing3: inputs, wing4: solenoids
        board3_config = b'\x22\x0d\x03\x03\x03\x07'     # wing1: lamps, wing2: lamps, wing3: lamps, wing4: hi-side lamps
        board4_config = b'\x23\x0d\x01\x01\x04\x05'     # wing1: sol, wing2: sol, wing3: matrix_out, wing4: matrix_in
        board1_version = b'\x20\x02\x02\x00\x00\x00'    # 2.0.0.0
        board2_version = b'\x21\x02\x02\x00\x00\x00'    # 2.0.0.0
        board3_version = b'\x22\x02\x02\x00\x00\x00'    # 2.0.0.0
        board4_version = b'\x23\x02\x02\x00\x00\x00'    # 2.0.0.0
        inputs1_message = b"\x20\x08\x00\xff\x00\x0c"   # inputs 0+1 off, 2+3 on, 8 on
        inputs2_message = b"\x21\x08\x00\x00\x00\x00"
        inputs3a_message = b"\x23\x08\x00\x00\x00\x00"
        inputs3b_message = b"\x23\x19\x00\x00\x00\x00\x00\x00\x00\x01"

        self.serialMock.expected_commands = {
            b'\xf0': b'\xf0\x20\x21\x22\x23',     # boards 20 + 21 + 22 + 23 installed
            self._crc_message(b'\x20\x0d\x00\x00\x00\x00'): self._crc_message(board1_config),
            self._crc_message(b'\x21\x0d\x00\x00\x00\x00'): self._crc_message(board2_config),
            self._crc_message(b'\x22\x0d\x00\x00\x00\x00'): self._crc_message(board3_config),
            self._crc_message(b'\x23\x0d\x00\x00\x00\x00'): self._crc_message(board4_config), # get config
            self._crc_message(b'\x20\x02\x00\x00\x00\x00'): self._crc_message(board1_version),
            self._crc_message(b'\x21\x02\x00\x00\x00\x00'): self._crc_message(board2_version),
            self._crc_message(b'\x22\x02\x00\x00\x00\x00'): self._crc_message(board3_version),
            self._crc_message(b'\x23\x02\x00\x00\x00\x00'): self._crc_message(board4_version),   # get version
            self._crc_message(b'\x20\x14\x00\x02\x17\x00'): False,  # configure coil 0
            self._crc_message(b'\x20\x14\x01\x04\x17\x00'): False,  # configure coil 1
            self._crc_message(b'\x20\x14\x02\x04\x0a\x00'): False,  # configure coil 2
            self._crc_message(b'\x20\x14\x03\x00\x0a\x06'): False,  # configure coil 3
            self._crc_message(b'\x21\x14\x0c\x00\x0a\x01'): False,  # configure coil 1-12
            self._crc_message(b'\x23\x14\x00\x02\x2a\x00'): False,  # configure coil 3-0
            self._crc_message(b'\x20\x13\x07\x00\x00\x00\x00', False): False,  # turn off all incands
            self._crc_message(b'\x22\x13\x07\x00\x00\x00\x00', False): False,  # turn off all incands
            self._crc_message(b'\x21\x40\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00\x00\x00'): False,  # turn off leds
        }
        self.serialMock.permanent_commands = {
            b'\xff': b'\xff',
            self._crc_message(b'\x20\x08\x00\x00\x00\x00'): self._crc_message(inputs1_message),
            self._crc_message(b'\x21\x08\x00\x00\x00\x00'): self._crc_message(inputs2_message),
            self._crc_message(b'\x23\x08\x00\x00\x00\x00'): self._crc_message(inputs3a_message),
            self._crc_message(b'\x23\x19\x00\x00\x00\x00\x00\x00\x00\x00'): self._crc_message(inputs3b_message),  # read inputs
        }
        super().setUp()

        assert isinstance(self.machine.default_platform, OppHardwarePlatform)

        self._wait_for_processing()
        self.assertEqual(0x02000000, self.machine.default_platform.min_version["com1"])

        self.assertFalse(self.serialMock.expected_commands)
        self.maxDiff = 100000

        # test hardware scan
        info_str = """Connected CPUs:
 - Port: com1 at 115200 baud. Chain Serial: com1
 -> Board: 0x20 Firmware: 0x2000000
 -> Board: 0x21 Firmware: 0x2000000
 -> Board: 0x22 Firmware: 0x2000000
 -> Board: 0x23 Firmware: 0x2000000

Incand cards:
 - Chain: com1 Board: 0x20 Card: 0 Numbers: [16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31]
 - Chain: com1 Board: 0x22 Card: 2 Numbers: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20,\
 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31]

Input cards:
 - Chain: com1 Board: 0x20 Card: 0 Numbers: [0, 1, 2, 3, 8, 9, 10, 11, 12, 13, 14, 15]
 - Chain: com1 Board: 0x21 Card: 1 Numbers: [0, 1, 2, 3, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21,\
 22, 23, 24, 25, 26, 27]
 - Chain: com1 Board: 0x23 Card: 3 Numbers: [0, 1, 2, 3, 8, 9, 10, 11]
 - Chain: com1 Board: 0x23 Card: 3 Numbers: [32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49,\
 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78,\
 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95]

Solenoid cards:
 - Chain: com1 Board: 0x20 Card: 0 Numbers: [0, 1, 2, 3]
 - Chain: com1 Board: 0x21 Card: 1 Numbers: [12, 13, 14, 15]
 - Chain: com1 Board: 0x23 Card: 3 Numbers: [0, 1, 2, 3, 4, 5, 6, 7]

LEDs:
 - Chain: com1 Board: 0x21 Card: 1
"""
        self.assertEqual(info_str, self.machine.default_platform.get_info_string())

    def testOpp(self):
        self._test_dual_wound_coils()
        self._test_switches()

    def _test_switches(self):
        # initial switches
        self.assertSwitchState("s_test", 1)
        self.assertSwitchState("s_test_no_debounce", 1)
        self.assertSwitchState("s_test_nc", 1)
        self.assertSwitchState("s_flipper", 0)
        self.assertSwitchState("s_test_card2", 1)
        self.assertSwitchState("s_matrix_test", 1)
        self.assertSwitchState("s_matrix_test2", 0)
        self.assertSwitchState("s_matrix_test3", 1)

        # switch change
        permanent_commands = copy.deepcopy(self.serialMock.permanent_commands)

        inputs1_message = b"\x20\x08\x00\x00\x01\x08"  # inputs 0+1+2 off, 3 on, 8 off
        inputs2_message = b'\x21\x08\x00\x00\x00\x00'
        inputs3a_message = b"\x23\x08\x00\x00\x00\x00"
        inputs3b_message = b"\x23\x19\x80\x00\x00\x00\x00\x01\x00\x00"
        self.serialMock.permanent_commands = {
            b'\xff': b'\xff',
            self._crc_message(b'\x20\x08\x00\x00\x00\x00'): self._crc_message(inputs1_message),
            self._crc_message(b'\x21\x08\x00\x00\x00\x00'): self._crc_message(inputs2_message),
            self._crc_message(b'\x23\x08\x00\x00\x00\x00'): self._crc_message(inputs3a_message),
            self._crc_message(b'\x23\x19\x00\x00\x00\x00\x00\x00\x00\x00'): self._crc_message(inputs3b_message),
        }

        switch = self.machine.switches["s_test_nc"]
        while self.machine.switch_controller.is_active(switch):
            self.advance_time_and_run(0.1)

        self.assertSwitchState("s_test", 1)
        self.assertSwitchState("s_test_no_debounce", 1)
        self.assertSwitchState("s_test_nc", 0)
        self.assertSwitchState("s_flipper", 0)
        self.assertSwitchState("s_test_card2", 0)
        self.assertSwitchState("s_matrix_test", 0)
        self.assertSwitchState("s_matrix_test2", 1)
        self.assertSwitchState("s_matrix_test3", 0)

        self.serialMock.permanent_commands = permanent_commands

    def _test_dual_wound_coils(self):
        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x02\x24\x0a\x00')] = False
        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x03\x23\x0a\x00')] = False
        self.serialMock.expected_commands[self._crc_message(b'\x20\x17\x03\x03')] = False
        self.serialMock.expected_commands[self._crc_message(b'\x20\x17\x03\x02')] = False
        self.machine.flippers["f_test_hold"].enable()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # enable a coil (when a rule is active)
        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x03\x21\x0a\x06')] = False
        self.serialMock.expected_commands[self._crc_message(b'\x20\x07\x00\x08\x00\x08', False)] = False
        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x03\x23\x0a\x00')] = False
        self.machine.coils["c_flipper_main"].enable()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # pulse it (when rule is active)
        self.serialMock.expected_commands[self._crc_message(b'\x20\x07\x00\x08\x00\x08', False)] = False
        self.machine.coils["c_flipper_main"].pulse()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # pulse it with other settings (when rule is active)
        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x03\x23\x2a\x00')] = False
        self.serialMock.expected_commands[self._crc_message(b'\x20\x07\x00\x08\x00\x08', False)] = False
        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x03\x23\x0a\x00')] = False
        self.machine.coils["c_flipper_main"].pulse(42)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x02\x04\x0a\x20')] = False
        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x03\x00\x0a\x26')] = False
        self.serialMock.expected_commands[self._crc_message(b'\x20\x17\x03\x83')] = False
        self.serialMock.expected_commands[self._crc_message(b'\x20\x17\x03\x82')] = False
        self.machine.flippers["f_test_hold"].disable()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # enable a coil (which is already configured right)
        self.serialMock.expected_commands[self._crc_message(b'\x20\x07\x00\x02\x00\x02', False)] = False
        self.machine.coils["c_test_allow_enable"].enable()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # disable it
        self.serialMock.expected_commands[self._crc_message(b'\x20\x07\x00\x00\x00\x02', False)] = False
        self.machine.coils["c_test_allow_enable"].disable()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # pulse it
        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x01\x02\x17\x00')] = False
        self.serialMock.expected_commands[self._crc_message(b'\x20\x07\x00\x02\x00\x02', False)] = False
        self.machine.coils["c_test_allow_enable"].pulse()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # pulse it again with same settings (no reconfigure)
        self.serialMock.expected_commands[self._crc_message(b'\x20\x07\x00\x02\x00\x02', False)] = False
        self.machine.coils["c_test_allow_enable"].pulse()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # pulse it with other settings (should reconfigure)
        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x01\x02\x2a\x00')] = False
        self.serialMock.expected_commands[self._crc_message(b'\x20\x07\x00\x02\x00\x02', False)] = False
        self.machine.coils["c_test_allow_enable"].pulse(42)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)


class TestOPP(OPPCommon, MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def setUp(self):
        self.expected_duration = 1.5
        opp.serial_imported = True
        opp.serial = MagicMock()
        self.serialMock = MockOppSocket("com1")
        board1_config = b'\x20\x0d\x01\x02\x03\x03'      # wing1: solenoids, wing2: inputs, wing3: lamps, wing4: lamps
        board2_config = b'\x21\x0d\x06\x02\x02\x01'      # wing1: neo, wing2: inputs, wing3: inputs, wing4: solenoids
        board1_version = b'\x20\x02\x00\x01\x01\x00'     # 0.1.1.0
        board2_version = b'\x21\x02\x00\x01\x01\x00'     # 0.1.1.0
        inputs1_message = b'\x20\x08\x00\x00\x00\x0c'    # inputs 0+1 off, 2+3 on, 8 on
        inputs2_message = b'\x21\x08\x00\x00\x00\x00'

        self.serialMock.expected_commands = {
            b'\xf0': b'\xf0\x20\x21',     # boards 20 + 21 installed
            self._crc_message(b'\x20\x0d\x00\x00\x00\x00'): self._crc_message(board1_config),
            self._crc_message(b'\x21\x0d\x00\x00\x00\x00'): self._crc_message(board2_config),     # get config
            self._crc_message(b'\x20\x02\x00\x00\x00\x00'): self._crc_message(board1_version),
            self._crc_message(b'\x21\x02\x00\x00\x00\x00'): self._crc_message(board2_version),   # get version
            self._crc_message(b'\x20\x14\x00\x02\x17\x00'): False,  # configure coil 0
            self._crc_message(b'\x20\x14\x01\x00\x17\x0f'): False,  # configure coil 1
            self._crc_message(b'\x20\x14\x02\x00\x0a\x0f'): False,  # configure coil 2
            self._crc_message(b'\x20\x14\x03\x00\x0a\x06'): False,  # configure coil 3
            self._crc_message(b'\x21\x14\x0c\x00\x0a\x01'): False,  # configure coil 1-12
            self._crc_message(b'\x20\x13\x07\x00\x00\x00\x00'): False,  # turn off all incands
            self._crc_message(b'\x21\x40\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00\x00\x00'): False   # turn off leds
        }
        self.serialMock.permanent_commands = {
            b'\xff': b'\xff',
            self._crc_message(b'\x20\x08\x00\x00\x00\x00'): self._crc_message(inputs1_message),
            self._crc_message(b'\x21\x08\x00\x00\x00\x00'): self._crc_message(inputs2_message),  # read inputs
        }
        super().setUp()

        self._wait_for_processing()

        self.assertFalse(self.serialMock.expected_commands)

    def test_opp(self):
        self._test_coils()
        self._test_leds()
        self._test_matrix_lights()
        self._test_autofires()
        self._test_switches()
        self._test_flippers()

        # test hardware scan
        self.maxDiff = 100000
        info_str = """Connected CPUs:
 - Port: com1 at 115200 baud. Chain Serial: com1
 -> Board: 0x20 Firmware: 0x10100
 -> Board: 0x21 Firmware: 0x10100

Incand cards:
 - Chain: com1 Board: 0x20 Card: 0 Numbers: [16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31]

Input cards:
 - Chain: com1 Board: 0x20 Card: 0 Numbers: [0, 1, 2, 3, 8, 9, 10, 11, 12, 13, 14, 15]
 - Chain: com1 Board: 0x21 Card: 1 Numbers: [0, 1, 2, 3, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27]

Solenoid cards:
 - Chain: com1 Board: 0x20 Card: 0 Numbers: [0, 1, 2, 3]
 - Chain: com1 Board: 0x21 Card: 1 Numbers: [12, 13, 14, 15]

LEDs:
 - Chain: com1 Board: 0x21 Card: 1
"""
        self.assertEqual(info_str, self.machine.default_platform.get_info_string())

    def _test_switches(self):
        # initial switches
        self.assertSwitchState("s_test", 1)
        self.assertSwitchState("s_test_no_debounce", 1)
        self.assertSwitchState("s_test_nc", 1)
        self.assertSwitchState("s_flipper", 0)
        self.assertSwitchState("s_test_card2", 1)

        # switch change
        permanent_commands = copy.deepcopy(self.serialMock.permanent_commands)

        inputs1_message = b"\x20\x08\x00\x00\x01\x08"  # inputs 0+1+2 off, 3 on, 8 off
        inputs2_message = b'\x21\x08\x00\x00\x00\x00'
        self.serialMock.permanent_commands = {
            b'\xff': b'\xff',
            self._crc_message(b'\x20\x08\x00\x00\x00\x00'): self._crc_message(inputs1_message),
            self._crc_message(b'\x21\x08\x00\x00\x00\x00'): self._crc_message(inputs2_message)
        }

        switch = self.machine.switches["s_test_nc"]
        while self.machine.switch_controller.is_active(switch):
            self.advance_time_and_run(0.1)

        self.assertSwitchState("s_test", 1)
        self.assertSwitchState("s_test_no_debounce", 1)
        self.assertSwitchState("s_test_nc", 0)
        self.assertSwitchState("s_flipper", 0)
        self.assertSwitchState("s_test_card2", 0)

        self.serialMock.permanent_commands = permanent_commands

    def _test_coils(self):
        self.assertEqual("OPP com1 Board 0x20", self.machine.coils["c_test"].hw_driver.get_board_name())
        # pulse coil
        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x00\x02\x17\x00')] = False   # configure coil 0
        self.serialMock.expected_commands[self._crc_message(b'\x20\x07\x00\x01\x00\x01')] = False
        self.machine.coils["c_test"].pulse()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands[self._crc_message(b'\x21\x14\x0c\x02\x0a\x00')] = False
        self.serialMock.expected_commands[self._crc_message(b'\x21\x07\x10\x00\x10\x00')] = False
        self.machine.coils["c_holdpower_16"].pulse(10)

        # enable coil (not allowed)
        with self.assertRaises(AssertionError):
            self.machine.coils["c_test"].enable()
        self._wait_for_processing()

        self.assertFalse(self.serialMock.expected_commands)
        self.assertFalse(self.serialMock.crashed)

        # disable coil
        self.serialMock.expected_commands[self._crc_message(b'\x20\x07\x00\x00\x00\x01', False)] = False
        self.machine.coils["c_test"].disable()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # pulse coil (with allow_enable set)
        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x01\x02\x17\x00')] = False
        self.serialMock.expected_commands[self._crc_message(b'\x20\x07\x00\x02\x00\x02', False)] = False
        self.machine.coils["c_test_allow_enable"].pulse()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # enable coil (with allow_enable set)
        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x01\x00\x17\x0f')] = False
        self.serialMock.expected_commands[self._crc_message(b'\x20\x07\x00\x02\x00\x02', False)] = False
        self.machine.coils["c_test_allow_enable"].enable()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

    def _test_matrix_lights(self):
        self.serialMock.expected_commands[self._crc_message(b'\x20\x13\x07\x00\x01\x00\x00', False)] = False
        self.machine.lights["test_light1"].on()
        self.machine.lights["test_light2"].off()

        self._wait_for_processing()

        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands[self._crc_message(b'\x20\x13\x07\x00\x03\x00\x00', False)] = False
        self.machine.lights["test_light1"].on()
        self.machine.lights["test_light2"].on()
        # it will only update once every 10 ticks so just advance 10 times to be sure
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

    def _test_leds(self):
        # set leds 0, 1, 2 to brightness 255
        self.serialMock.expected_commands[self._crc_message(b'\x21\x40\x00\x00\x00\x03\x00\x00\xff\xff\xff', False)] = False

        self.machine.lights["test_led1"].on()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # set leds 0, 1, 2 to brightness 0
        # set leds 3, 4, 5 to brightness 255
        self.serialMock.expected_commands[self._crc_message(b'\x21\x40\x00\x00\x00\x06\x00\x00\x00\x00\x00\xff\xff\xff', False)] = False

        self.machine.lights["test_led1"].off()
        self.machine.lights["test_led2"].on()

        self._wait_for_processing()

        self.assertFalse(self.serialMock.expected_commands)

        # align with update task
        self.advance_time_and_run(.1)

        # two fades which are close enough together are batched
        self.serialMock.expected_commands[self._crc_message(b'\x21\x40\x00\x00\x00\x06\x00\x64\xff\x00\x00\xff\x00\x00', False)] = False
        self.machine.lights["test_led1"].color("red", fade_ms=100)
        self.machine.lights["test_led2"].color("red", fade_ms=95)

        # align with update task
        self.advance_time_and_run(.1)

        # fade leds 3, 4, 5 to brightness 245, 222, 179
        self.serialMock.expected_commands[self._crc_message(b'\x21\x40\x00\x03\x00\x03\x07\xd0\xf5\xde\xb3', False)] = False

        self.machine.lights["test_led2"].color("wheat", fade_ms=2000)

        self._wait_for_processing()

        self.assertFalse(self.serialMock.expected_commands)

    def _test_autofires(self):
        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x00\x03\x17\x20')] = False
        self.machine.autofires["ac_slingshot_test"].enable()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x00\x02\x17\x20')] = False
        self.machine.autofires["ac_slingshot_test"].disable()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x01\x03\x17\x30')] = False
        self.machine.autofires["ac_slingshot_test2"].enable()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x01\x00\x17\x3f')] = False
        self.machine.autofires["ac_slingshot_test2"].disable()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x00\x0b\x17\x14')] = False
        self.machine.autofires["ac_delayed_kickback"].enable()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x00\x02\x17\x20')] = False
        self.machine.autofires["ac_delayed_kickback"].disable()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

    def _test_flippers(self):
        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x03\x21\x0a\x06')] = False
        self.machine.flippers["f_test_single"].enable()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands[self._crc_message(b'\x20\x14\x03\x00\x0a\x26')] = False
        self.machine.flippers["f_test_single"].disable()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)



"""Plugin config player."""
from mpf.config_players.device_config_player import DeviceConfigPlayer


class PluginPlayer(DeviceConfigPlayer):

    """Base class for a remote ConfigPlayer that is registered as a plug-in to MPF.

    This class is created on the MPF side of things.
    """

    __slots__ = ["bcp_client", "_show_keys"]

    def __init__(self, machine):
        """Initialise plugin player."""
        super().__init__(machine)
        self.bcp_client = None
        self._show_keys = {}

    def __repr__(self):
        """Return str representation."""
        return 'PluginPlayer.{}'.format(self.show_section)

    def get_express_config(self, value):
        """Not supported."""
        del value
        raise AssertionError("{} does not support express config".format(self))

    def _get_bcp_client(self, config):
        client_name = config.get('bcp_connection', "local_display")
        client = self.machine.bcp.transport.get_named_client(client_name)
        if not client:
            raise AssertionError(
                "bcp connection {} not found".format(client_name))

        return client

    def _add_handlers(self):
        self.machine.events.add_handler('init_phase_1',
                                        self._initialize_mode_handlers, priority=20)
        # since bcp is connecting in init_phase_2 we have to postpone this
        self.machine.events.add_handler('init_phase_3',
                                        self._initialise_system_wide)

    def register_player_events(self, config, mode=None, priority=0):
        """Register player events via BCP.

        Override this method in the base class and registers the
        config_player events to send the trigger via BCP instead of calling
        the local play() method.
        """
        events = super().register_player_events(config, mode, priority)
        # when bcp is disabled do not register plugin_player
        if not self.machine.options['bcp']:
            return events

        self.bcp_client = self._get_bcp_client(config)

        self.machine.bcp.interface.add_registered_trigger_event_for_client(
            self.bcp_client, '{}_play'.format(self.show_section))
        self.machine.bcp.interface.add_registered_trigger_event_for_client(
            self.bcp_client, '{}_clear'.format(self.show_section))

        return events

    # pylint: disable-msg=too-many-arguments
    def show_play_callback(self, settings, priority, calling_context, show_tokens, context, start_time):
        """Register BCP events."""
        config = {'bcp_connection': settings['bcp_connection']} if 'bcp_connection' in settings else {}
        event_keys = self.register_player_events(config, None, priority)
        self._show_keys[context + self.config_file_section] = event_keys
        super().show_play_callback(settings, priority, calling_context, show_tokens, context, start_time)

    def show_stop_callback(self, context):
        """Remove BCP events."""
        self.unload_player_events(self._show_keys[context + self.config_file_section])
        del self._show_keys[context + self.config_file_section]
        super().show_stop_callback(context)

    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Trigger remote player via BCP."""
        self.machine.bcp.interface.bcp_trigger(
            name='{}_play'.format(self.show_section),
            settings=settings, context=context, calling_context=calling_context,
            priority=priority, **kwargs)

    def clear_context(self, context):
        """Clear the context at remote player via BCP."""
        self.machine.bcp.interface.bcp_trigger(
            name='{}_clear'.format(self.show_section),
            context=context)



from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase
from unittest.mock import MagicMock

from mpf.tests.MpfTestCase import MpfTestCase


class TestDropTargets(MpfTestCase):

    def get_config_file(self):
        return 'test_drop_targets.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/drop_targets/'

    def get_platform(self):
        return 'smart_virtual'

    def test_reset_and_playfield_active(self):
        self.mock_event("playfield_active")
        self.hit_switch_and_run("switch1", 1)
        self.hit_switch_and_run("switch2", 1)

        # playfield should is active when drop target are shot down
        self.assertEventCalled("playfield_active")
        self.mock_event("playfield_active")

        self.assertTrue(self.machine.drop_targets["left1"].complete)
        self.assertTrue(self.machine.drop_targets["left2"].complete)
        self.assertFalse(self.machine.drop_targets["left3"].complete)

        # reset the bank. this should not trigger playfield_active
        self.machine.drop_target_banks["left_bank"].reset()
        self.advance_time_and_run(1)

        self.assertEventNotCalled("playfield_active")
        self.assertFalse(self.machine.drop_targets["left1"].complete)
        self.assertFalse(self.machine.drop_targets["left2"].complete)
        self.assertFalse(self.machine.drop_targets["left3"].complete)

    def test_drop_target_bank(self):
        self.assertIn('left1', self.machine.drop_targets)
        self.assertIn('left2', self.machine.drop_targets)
        self.assertIn('left3', self.machine.drop_targets)
        self.assertIn('left_bank', self.machine.drop_target_banks)

        self.machine.coils["coil1"].pulse = MagicMock()

        self.assertFalse(self.machine.drop_targets["left1"].complete)
        self.assertFalse(self.machine.drop_targets["left2"].complete)
        self.assertFalse(self.machine.drop_targets["left3"].complete)
        self.assertFalse(self.machine.drop_target_banks["left_bank"].complete)

        self.hit_switch_and_run("switch1", 1)
        self.hit_switch_and_run("switch2", 1)
        self.assertTrue(self.machine.drop_targets["left1"].complete)
        self.assertTrue(self.machine.drop_targets["left2"].complete)
        self.assertFalse(self.machine.drop_targets["left3"].complete)
        self.assertFalse(self.machine.drop_target_banks["left_bank"].complete)

        assert not self.machine.coils["coil1"].pulse.called

        self.hit_switch_and_run("switch3", .5)
        self.assertTrue(self.machine.drop_targets["left1"].complete)
        self.assertTrue(self.machine.drop_targets["left2"].complete)
        self.assertTrue(self.machine.drop_targets["left3"].complete)
        self.assertTrue(self.machine.drop_target_banks["left_bank"].complete)

        assert not self.machine.coils["coil1"].pulse.called

        # it should reset after 1s
        self.advance_time_and_run(.5)
        self.machine.coils["coil1"].pulse.assert_called_once_with(max_wait_ms=100)

        # after another 100ms the switches releases
        self.release_switch_and_run("switch1", 0)
        self.release_switch_and_run("switch2", 0)
        self.release_switch_and_run("switch3", 1)

        self.assertFalse(self.machine.drop_targets["left1"].complete)
        self.assertFalse(self.machine.drop_targets["left2"].complete)
        self.assertFalse(self.machine.drop_targets["left3"].complete)
        self.assertFalse(self.machine.drop_target_banks["left_bank"].complete)

        # check that the bank does not reset if already down
        self.machine.coils["coil1"].pulse = MagicMock()
        self.machine.drop_target_banks['left_bank'].reset()
        assert not self.machine.coils["coil1"].pulse.called

        # reset should work with one target down
        self.hit_switch_and_run("switch1", 1)
        self.machine.drop_target_banks['left_bank'].reset()
        self.machine.coils["coil1"].pulse.assert_called_once_with(max_wait_ms=100)

    def test_knockdown_and_reset(self):
        self.mock_event("unexpected_ball_on_playfield")
        self.machine.coils["coil2"].pulse = MagicMock(wraps=self.machine.coils["coil2"].pulse)
        self.machine.coils["coil3"].pulse = MagicMock(wraps=self.machine.coils["coil3"].pulse)

        self.assertFalse(self.machine.drop_targets["left6"].complete)

        # knock it down
        self.post_event("knock_knock")
        self.advance_time_and_run(.3)
        assert not self.machine.coils["coil2"].pulse.called
        self.machine.coils["coil3"].pulse.assert_called_once_with(max_wait_ms=100)

        # ignore ms means the state is not updated yet
        self.assertFalse(self.machine.drop_targets["left6"].complete)
        self.advance_time_and_run(.3)

        # and now it is
        self.assertTrue(self.machine.drop_targets["left6"].complete)

        # reset it
        self.machine.coils["coil3"].pulse.reset_mock()

        self.post_event("reset_target")
        self.advance_time_and_run(.3)
        assert not self.machine.coils["coil3"].pulse.called
        self.machine.coils["coil2"].pulse.assert_called_once_with(max_wait_ms=100)

        # ignore ms means the state is not updated yet
        self.assertTrue(self.machine.drop_targets["left6"].complete)
        self.advance_time_and_run(6)

        # and now it is
        self.assertFalse(self.machine.drop_targets["left6"].complete)

        self.assertEventNotCalled("unexpected_ball_on_playfield")

    def test_drop_targets_in_mode(self):
        self.machine.modes['mode1'].start()
        self.advance_time_and_run()

        self.machine.coils["coil2"].pulse = MagicMock()

        self.assertFalse(self.machine.drop_targets["left4"].complete)
        self.assertFalse(self.machine.drop_targets["left5"].complete)
        self.assertFalse(self.machine.drop_targets["left6"].complete)
        self.assertFalse(self.machine.drop_target_banks["left_bank_2"].complete)

        self.hit_switch_and_run("switch4", 1)
        self.hit_switch_and_run("switch5", 1)
        self.assertTrue(self.machine.drop_targets["left4"].complete)
        self.assertTrue(self.machine.drop_targets["left5"].complete)
        self.assertFalse(self.machine.drop_targets["left6"].complete)
        self.assertFalse(self.machine.drop_target_banks["left_bank_2"].complete)

        self.machine.modes['mode1'].stop()
        self.advance_time_and_run()

        self.assertTrue(self.machine.drop_targets["left4"].complete)
        self.assertTrue(self.machine.drop_targets["left5"].complete)
        self.assertFalse(self.machine.drop_targets["left6"].complete)
        self.assertFalse(self.machine.drop_target_banks["left_bank_2"].complete)

        # should not complete the bank
        self.hit_switch_and_run("switch6", .1)

        self.assertTrue(self.machine.drop_targets["left4"].complete)
        self.assertTrue(self.machine.drop_targets["left5"].complete)
        self.assertTrue(self.machine.drop_targets["left6"].complete)

        self.assertFalse(self.machine.drop_target_banks["left_bank_2"].complete)

        self.post_event("reset_target")

        self.machine.modes['mode1'].start()
        self.advance_time_and_run()

        # mode is running again. should complete
        self.hit_switch_and_run("switch4", .1)
        self.hit_switch_and_run("switch5", .1)
        self.hit_switch_and_run("switch6", .1)

        self.assertTrue(self.machine.drop_targets["left4"].complete)
        self.assertTrue(self.machine.drop_targets["left5"].complete)
        self.assertTrue(self.machine.drop_targets["left6"].complete)

        self.assertTrue(self.machine.drop_target_banks["left_bank_2"].complete)

    def test_drop_target_reset(self):
        target = self.machine.drop_targets["left6"]
        self.machine.coils["coil2"].pulse = MagicMock()
        self.machine.coils["coil3"].pulse = MagicMock()
        self.assertSwitchState("switch6", 0)

        # target up. it should not reset
        target.reset()
        self.advance_time_and_run()

        assert not self.machine.coils["coil2"].pulse.called
        assert not self.machine.coils["coil3"].pulse.called

        # hit target down
        self.hit_switch_and_run("switch6", 1)
        self.assertTrue(target.complete)

        # it should reset
        target.reset()
        self.advance_time_and_run()

        self.machine.coils["coil2"].pulse.assert_called_once_with(max_wait_ms=100)
        self.machine.coils["coil2"].pulse.reset_mock()
        assert not self.machine.coils["coil3"].pulse.called
        self.release_switch_and_run("switch6", 1)

        # knock down should work
        target.knockdown()
        self.advance_time_and_run()

        self.machine.coils["coil3"].pulse.assert_called_once_with(max_wait_ms=100)
        self.machine.coils["coil3"].pulse.reset_mock()
        assert not self.machine.coils["coil2"].pulse.called
        self.hit_switch_and_run("switch6", 1)

        # but not when its down already
        target.knockdown()
        self.advance_time_and_run()
        assert not self.machine.coils["coil2"].pulse.called
        assert not self.machine.coils["coil3"].pulse.called

    def test_drop_target_ignore_ms(self):

        self.mock_event('drop_target_center1_down')
        self.mock_event('drop_target_center1_up')

        self.hit_switch_and_run('switch10', 1)
        self.assertSwitchState('switch10', True)  # ###############

        self.assertEventNotCalled('drop_target_center1_up')
        self.assertEventCalled('drop_target_center1_down')

        self.post_event('reset_center1', .05)
        self.release_switch_and_run('switch10', .1)
        self.hit_switch_and_run('switch10', .1)
        self.release_switch_and_run('switch10', .1)
        self.assertSwitchState('switch10', False)

        self.advance_time_and_run(.5)

        # reset happened in the ignore window so this event should not be
        # called
        self.assertEventNotCalled('drop_target_center1_up')
        self.advance_time_and_run(1)

        # now do the same test for knockdown

        self.mock_event('drop_target_center1_down')

        self.post_event('knockdown_center1', .2)
        self.hit_switch_and_run('switch10', .1)
        self.assertEventNotCalled('drop_target_center1_down')

        self.advance_time_and_run(1)
        self.assertEventCalled('drop_target_center1_down')

    def test_drop_target_bank_ignore_ms(self):
        self.mock_event('drop_target_bank_right_bank_down')
        self.mock_event('drop_target_bank_right_bank_mixed')

        self.hit_switch_and_run('switch8', 1)
        self.hit_switch_and_run('switch9', 1)

        self.assertEventCalled('drop_target_bank_right_bank_mixed', 1)
        self.assertEventCalled('drop_target_bank_right_bank_down', 1)

        self.mock_event('drop_target_bank_right_bank_down')
        self.mock_event('drop_target_bank_right_bank_mixed')
        self.mock_event('drop_target_bank_right_bank_up')

        self.post_event('reset_right_bank', .5)

        # these events should not be called since we're in the ignore window
        self.assertEventNotCalled('drop_target_bank_right_bank_mixed')
        self.assertEventNotCalled('drop_target_bank_right_bank_up')

        self.advance_time_and_run(1)

        # after 1s, the ignore is cleared and the bank updates its state.

        # mixed should not have been called since it happened during the
        # ignore window
        self.assertEventNotCalled('drop_target_bank_right_bank_mixed')

        # up should have been called by now
        self.assertEventCalled('drop_target_bank_right_bank_up')


class TestDropTargetsInGame(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'test_multiple_drop_resets_on_startup.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/drop_targets/'

    def test_multiple_reset_events(self):
        """Check that the drop target tries to reset three times on startup."""
        # we check that the bank resets 3 times but only reports down once
        self.mock_event("drop_target_bank_multiple_resets_on_game_start_down")
        self.machine.coils["coil1"].hw_driver.pulse = MagicMock()

        # drop is down
        self.hit_switch_and_run("switch1", 1)
        self.assertEventCalled("drop_target_bank_multiple_resets_on_game_start_down", times=1)
        self.assertEqual(0, self.machine.coils["coil1"].hw_driver.pulse.call_count)

        # start game to trigger the reset
        self.start_game()
        # drop should have tried a reset
        self.assertEventCalled("drop_target_bank_multiple_resets_on_game_start_down", times=1)
        self.assertEqual(1, self.machine.coils["coil1"].hw_driver.pulse.call_count)

        # it should reset again
        self.advance_time_and_run(3)
        self.assertEventCalled("drop_target_bank_multiple_resets_on_game_start_down", times=1)
        self.assertEqual(2, self.machine.coils["coil1"].hw_driver.pulse.call_count)

        # it should reset again
        self.advance_time_and_run(3)
        self.assertEventCalled("drop_target_bank_multiple_resets_on_game_start_down", times=1)
        self.assertEqual(3, self.machine.coils["coil1"].hw_driver.pulse.call_count)



"""Test Switch Position Mixin."""
from mpf.core.rgb_color import RGBColor
from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase


class TestSwitchPositions(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/switch_player/'

    def test_switch_positions(self):
        switch1 = self.machine.switches["s_test1"]
        switch2 = self.machine.switches["s_test2"]
        switch3 = self.machine.switches["s_test3"]

        self.assertEqual(switch1.x, 0.4)
        self.assertEqual(switch1.y, 0.5)
        self.assertEqual(switch1.z, 0)

        self.assertEqual(switch2.x, 0.6)
        self.assertEqual(switch2.y, 0.7)
        self.assertEqual(switch2.z, None)

        self.assertEqual(switch3.x, None)
        self.assertEqual(switch3.y, None)
        self.assertEqual(switch3.z, None)



"""Baseclass for ball device ejectors."""

MYPY = False
if MYPY:    # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.ball_device.ball_device import BallDevice  # pylint: disable-msg=cyclic-import,unused-import


class BallDeviceEjector:

    """Ejector for a ball device.

    It has to implement at least one of eject_one_ball or eject_all_balls.
    """

    __slots__ = ["config", "ball_device", "machine"]

    def __init__(self, config: dict, ball_device: "BallDevice", machine: "MachineController") -> None:
        """Initialise ejector."""
        self.config = config
        self.ball_device = ball_device
        self.machine = machine

    async def eject_one_ball(self, is_jammed, eject_try, balls_in_device):
        """Eject one ball."""
        raise NotImplementedError()

    async def reorder_balls(self):
        """Reorder balls without ejecting.

        This might be useful when count become unstable during a jam condition.
        """
        raise NotImplementedError()

    def ball_search(self, phase, iteration):
        """Search ball in device."""
        raise NotImplementedError()



"""Handles outgoing balls."""
import asyncio

from typing import List, Optional

from mpf.core.utility_functions import Util
from mpf.devices.ball_device.ball_count_handler import EjectTracker
from mpf.devices.ball_device.ball_device_state_handler import BallDeviceStateHandler
from mpf.devices.ball_device.incoming_balls_handler import IncomingBall

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.devices.ball_device.ball_device import BallDevice  # pylint: disable-msg=cyclic-import,unused-import


class OutgoingBall:

    """One outgoing ball."""

    __slots__ = ["max_tries", "eject_timeout", "target", "player_controlled", "already_left"]

    def __init__(self, target: "BallDevice") -> None:
        """Initialise outgoing ball."""
        self.max_tries = 0                  # type: int
        self.eject_timeout = 0              # type: int
        self.target = target                # type: BallDevice
        self.player_controlled = False      # type: bool
        self.already_left = False           # type: bool


class OutgoingBallsHandler(BallDeviceStateHandler):

    """Handles all outgoing balls."""

    __slots__ = ["_eject_queue", "_current_target", "_cancel_future", "_incoming_ball_which_may_skip",
                 "_no_incoming_ball_which_may_skip", "_incoming_ball_which_may_skip_obj", "_eject_future"]

    def __init__(self, ball_device: "BallDevice") -> None:
        """Initialise outgoing balls handler."""
        super().__init__(ball_device)
        self._eject_queue = asyncio.Queue()     # type: asyncio.Queue
        self._current_target = None     # type: Optional[BallDevice]
        self._cancel_future = None      # type: Optional[asyncio.Future]
        self._incoming_ball_which_may_skip = asyncio.Event()
        self._incoming_ball_which_may_skip.clear()
        self._no_incoming_ball_which_may_skip = asyncio.Event()
        self._no_incoming_ball_which_may_skip.set()
        self._incoming_ball_which_may_skip_obj = []     # type: List[IncomingBall]
        self._eject_future = None       # type: Optional[asyncio.Future]

    def add_eject_to_queue(self, eject: OutgoingBall):
        """Add an eject request to queue."""
        self._eject_queue.put_nowait(eject)

    def add_incoming_ball_which_may_skip(self, incoming_ball: IncomingBall):
        """Add incoming ball which may skip the device."""
        self._incoming_ball_which_may_skip_obj.append(incoming_ball)
        self._no_incoming_ball_which_may_skip.clear()
        self._incoming_ball_which_may_skip.set()

    def remove_incoming_ball_which_may_skip(self, incoming_ball: IncomingBall):
        """Remove incoming ball which may skip the device."""
        try:
            self._incoming_ball_which_may_skip_obj.remove(incoming_ball)
        except ValueError as e:
            if self.unit_test:
                # re-raise this in tests
                raise e
            self.warning_log("Double remove of incoming ball. This is likely a bug! "
                             "Please report in the forum if you can reproduce it.")
        if not self._incoming_ball_which_may_skip_obj:
            self._incoming_ball_which_may_skip.clear()
            self._no_incoming_ball_which_may_skip.set()

    @property
    def is_idle(self):
        """Return true if idle."""
        return not self._current_target and self._eject_queue.empty()

    @property
    def is_ready_to_receive(self):
        """Return true if we can receive balls."""
        return not self._current_target or not self._current_target.is_playfield() or not self._eject_future

    async def wait_for_ready_to_receive(self):
        """Wait until the outgoing balls handler is ready to receive."""
        # if we are ejecting to a playfield wait until the eject finished because we cannot properly confirm otherwise
        if not self.is_ready_to_receive:
            self.debug_log("Wait for eject to finish")
            await self._eject_future
            self.debug_log("Eject finished")

    async def _run(self):
        """Wait for eject queue."""
        while True:
            self._current_target = None
            self.ball_device.set_eject_state("idle")
            self.debug_log("Waiting for eject request.")
            eject_queue_future = asyncio.ensure_future(self._eject_queue.get())
            incoming_ball_which_may_skip = self._incoming_ball_which_may_skip.wait()
            event = await Util.first([eject_queue_future, incoming_ball_which_may_skip])

            if event == eject_queue_future:
                eject_request = await event
                self._current_target = eject_request.target
                self.debug_log("Got eject request")

                if eject_request.already_left:
                    ball_eject_process = await self.ball_device.ball_count_handler.start_eject(already_left=True)
                    # no prepare eject because this cannot be blocked
                    await self._post_ejecting_event(eject_request, 1)
                    incoming_ball_at_target = self._add_incoming_ball_to_target(eject_request.target)
                    result = await self._handle_confirm(eject_request, ball_eject_process,
                                                        incoming_ball_at_target, 1)
                    if result:
                        await self.ball_device.ball_count_handler.end_eject(ball_eject_process, True)
                        continue

                if not await self._ejecting(eject_request):
                    return
            else:
                await self._skipping_ball(self.ball_device.config['eject_targets'][0], True)

    async def _skipping_ball(self, target: "BallDevice", add_ball_to_target: bool):
        incoming_skipping_ball = self._incoming_ball_which_may_skip_obj[0]
        self.info_log("Expecting incoming ball which may skip the device.")
        eject_request = OutgoingBall(target)
        await self._post_ejecting_event(eject_request, 1)
        incoming_ball_at_target = self._add_incoming_ball_to_target(eject_request.target)
        confirm_future = asyncio.ensure_future(incoming_ball_at_target.wait_for_confirm())
        ball_future = asyncio.ensure_future(self.ball_device.ball_count_handler.wait_for_ball())
        no_incoming_future = asyncio.ensure_future(self._no_incoming_ball_which_may_skip.wait())
        futures = [confirm_future, no_incoming_future, ball_future]
        if self._cancel_future:
            futures.append(self._cancel_future)

        if target.is_playfield():
            timeout = self.ball_device.config['eject_timeouts'][target] / 1000
        else:
            timeout = None

        has_timeouted = False
        try:
            await Util.first(futures, timeout=timeout)
        except asyncio.TimeoutError:
            has_timeouted = True

        # if we got an confirm
        if (confirm_future.done() and not confirm_future.cancelled()) or has_timeouted:
            self.info_log("Got confirm for skipping ball.")
            await self._handle_eject_success(eject_request)
            incoming_skipping_ball.ball_arrived()
            if add_ball_to_target:
                target.available_balls += 1
            return True

        target.remove_incoming_ball(incoming_ball_at_target)
        await self._failed_eject(eject_request, 1, True)

        self.info_log("No longer expecting incoming ball which may skip the device.")
        return False

    def find_available_ball_in_path(self, start: "BallDevice") -> bool:
        """Try to remove available ball at the end of the path."""
        if self._current_target == start:
            self.debug_log("Loop detected. Path will not go anywhere.")
            return False

        if not self._current_target and self.ball_device.available_balls > 0:
            self.debug_log("We do not have an eject but an available ball.")
            return True

        if self._current_target:
            if self._current_target.is_playfield():
                self.debug_log("End of path is playfield %s", self._current_target)
                return True

            return self._current_target.find_available_ball_in_path(start)

        self.ball_device.log.warning("No eject and no available_balls. Path went nowhere.")
        return False

    def cancel_path_if_target_is(self, start: "BallDevice", target: "BallDevice") -> bool:
        """Check if the ball is going to a certain target and cancel the path in that case.

        Args:
        ----
            start: Start ball device.
            target: Target to check

        Returns: True if found and deleted.
        """
        if self._current_target == start:
            self.debug_log("Loop detected. Path will not go anywhere.")
            return False

        # TODO: check queue entries
        if not self._cancel_future or self._cancel_future.done():
            # we cannot cancel anyway so do not even check further
            self.debug_log("Cancel path if target is not %s failed. Cannot cancel eject.", target.name)
            return False

        if not self._current_target:
            # no current target -> success we are not ejecting to the target
            self.debug_log("Cancel path if target is not %s failed. No current target.", target.name)
            return False

        if self._current_target == target:
            self.debug_log("Cancel path if target is not %s successful.", target.name)
            target.available_balls -= 1
            self._cancel_future.set_result(True)
            return True

        if not self._current_target.is_playfield() and self._current_target.cancel_path_if_target_is(start, target):
            # our successors are ejecting to target. cancel eject
            self.debug_log("Cancel path if target is not %s successful at successors.", target.name)
            self._cancel_future.set_result(True)
            return True

        # default false
        self.debug_log("Cancel path if target is not %s failed. We got another target.", target.name)
        return False

    # pylint: disable-msg=inconsistent-return-statements
    async def _ejecting(self, eject_request: OutgoingBall):
        """Perform main eject loop."""
        assert self._current_target is not None
        eject_try = 0
        while True:
            # make sure the count is currently valid. process incoming and lost balls
            await self.ball_device.ball_count_handler.wait_for_count_is_valid()

            # prevent physical races with eject confirm
            if self._current_target.is_playfield() and not self.ball_device.ball_count_handler.is_full:
                await self.ball_device.incoming_balls_handler.wait_for_no_incoming_balls()

            if not self.ball_device.ball_count_handler.has_ball:
                # wait until we have a ball
                self._cancel_future = asyncio.Future()
                ball_future = asyncio.ensure_future(self.ball_device.ball_count_handler.wait_for_ball())
                skipping_ball_future = asyncio.ensure_future(self._incoming_ball_which_may_skip.wait())

                self.ball_device.set_eject_state("waiting_for_ball")
                result = await Util.first([self._cancel_future, ball_future, skipping_ball_future])

                if result == skipping_ball_future:
                    self._cancel_future = asyncio.Future()
                    result = await self._skipping_ball(self._current_target, False)
                    if result or self._cancel_future.done() and not self._cancel_future.cancelled():
                        self._cancel_future = None
                        return True

                    self._cancel_future.cancel()
                    self._cancel_future = None
                    continue

                if self._cancel_future.done() and not self._cancel_future.cancelled():
                    # eject cancelled
                    self._cancel_future = None
                    return True
                self._cancel_future.cancel()
                self._cancel_future = None

            self.ball_device.set_eject_state("waiting_for_target_ready")

            # inform targets about the eject (can delay the eject)
            await self._prepare_eject(eject_request, eject_try)
            # wait for target to be ready
            # TODO: block one spot in target device to prevent double eject
            await eject_request.target.wait_for_ready_to_receive(self.ball_device)
            self.ball_device.set_eject_state("ejecting")
            self._eject_future = asyncio.Future()
            result = await self._eject_ball(eject_request, eject_try)
            self._eject_future.set_result(result)
            self._eject_future = None
            if result:
                # eject is done. return to main loop
                return True

            eject_try += 1

            if eject_request.max_tries and eject_try >= eject_request.max_tries:
                # stop device
                self.ball_device.set_eject_state("eject_broken")
                await self._failed_eject(eject_request, eject_try, False)
                self.machine.events.post("balldevice_{}_broken".format(self.ball_device.name))
                '''event: balldevice_(name)_broken
                config_section: ball_devices
                class_label: ball_device

                desc: The ball device called "name" is broken and will no longer operate.
                '''
                self._task.cancel()
                return False

            await self._failed_eject(eject_request, eject_try, True)

    async def _prepare_eject(self, eject_request: OutgoingBall, eject_try: int):
        await self.machine.events.post_queue_async(
            'balldevice_{}_ball_eject_attempt'.format(self.ball_device.name),
            balls=1,
            target=eject_request.target,
            source=self.ball_device,
            mechanical_eject=eject_request.player_controlled,
            num_attempts=eject_try)
        '''event: balldevice_(name)_ball_eject_attempt
        config_section: ball_devices
        class_label: ball_device

        desc: The ball device called "name" is attempting to eject a ball (or
        balls). This is a queue event. The eject will not actually be attempted
        until the queue is cleared.

        args:

        balls: The number of balls that are to be ejected.
        target: The target ball device that will receive these balls.
        source: The source device that will be ejecting the balls.
        mechanical_eject: Boolean as to whether this is a mechanical eject.
        num_attempts: How many eject attempts have been tried so far.
        '''

    async def _failed_eject(self, eject_request: OutgoingBall, eject_try: int, retry: bool):
        await self.machine.events.post_async(
            'balldevice_' + self.ball_device.name + '_ball_eject_failed',
            target=eject_request.target,
            balls=1,
            retry=retry,
            num_attempts=eject_try)
        '''event: balldevice_(name)_ball_eject_failed
        config_section: ball_devices
        class_label: ball_device

        desc: A ball (or balls) has failed to eject from the device (name).
        args:
            target: The target device that was supposed to receive the ejected
                balls.
            balls: The number of balls that failed to eject.
            retry: Boolean as to whether this eject will be retried.
            num_attempts: How many attemps have been made to eject this ball
                (or balls).
        '''

    async def _post_ejecting_event(self, eject_request: OutgoingBall, eject_try: int):
        await self.machine.events.post_async(
            'balldevice_{}_ejecting_ball'.format(self.ball_device.name),
            balls=1,
            target=eject_request.target,
            source=self.ball_device,
            mechanical_eject=eject_request.player_controlled,
            num_attempts=eject_try)
        '''event: balldevice_(name)_ejecting_ball
        config_section: ball_devices
        class_label: ball_device

        desc: The ball device called "name" is ejecting a ball right now.

        args:

        balls: The number of balls that are to be ejected.
        target: The target ball device that will receive these balls.
        source: The source device that will be ejecting the balls.
        mechanical_eject: Boolean as to whether this is a mechanical eject.
        num_attempts: How many eject attempts have been tried so far.
        '''

    async def _eject_ball(self, eject_request: OutgoingBall, eject_try: int) -> bool:
        # inform the counter that we are ejecting now
        self.info_log("Ejecting ball to %s", eject_request.target)
        await self._post_ejecting_event(eject_request, eject_try)
        ball_eject_process = await self.ball_device.ball_count_handler.start_eject()
        try:
            await ball_eject_process.will_eject()
            self.info_log("Wait for ball to leave device")
            # eject the ball

            ball_left = ball_eject_process.wait_for_ball_left()
            waiters = [ball_left]
            trigger = None
            tilt = None
            if self.ball_device.ejector:
                # eject on tilt
                if eject_request.player_controlled:
                    tilt = self.machine.events.wait_for_event("tilt")
                    waiters.append(tilt)

                # wait for trigger event
                if eject_request.player_controlled and self.ball_device.config['player_controlled_eject_event']:
                    trigger = self.machine.events.wait_for_event(
                        self.ball_device.config['player_controlled_eject_event'])
                    waiters.append(trigger)
                elif eject_request.player_controlled and self.ball_device.config['mechanical_eject']:
                    # do nothing
                    pass
                else:
                    await self.ball_device.ejector.eject_one_ball(ball_eject_process.is_jammed(), eject_try,
                                                                  self.ball_device.ball_count_handler.handled_balls)

            # wait until the ball has left
            if (self.ball_device.config['mechanical_eject'] or
                    self.ball_device.config['player_controlled_eject_event']) and eject_request.player_controlled:
                timeout = None
            else:
                timeout = eject_request.eject_timeout
            try:
                await Util.any(waiters, timeout=timeout)
            except asyncio.TimeoutError:
                # timeout. ball did not leave. failed
                await self.ball_device.ball_count_handler.end_eject(ball_eject_process, False)
                return False

            if (trigger and trigger.done()) or (tilt and tilt.done()):
                await self.ball_device.ejector.eject_one_ball(ball_eject_process.is_jammed(), eject_try,
                                                              self.ball_device.ball_count_handler.handled_balls)
                # TODO: add timeout here
                await ball_left

            self.ball_device.set_eject_state("ball_left")
            self.info_log("Ball left")
            incoming_ball_at_target = self._add_incoming_ball_to_target(eject_request.target)
            result = await self._handle_confirm(eject_request, ball_eject_process, incoming_ball_at_target,
                                                eject_try)
            await self.ball_device.ball_count_handler.end_eject(ball_eject_process, result)
            return result
        except asyncio.CancelledError:
            ball_eject_process.cancel()
            raise

    def _add_incoming_ball_to_target(self, target: "BallDevice") -> IncomingBall:
        # we are the source of this ball
        incoming_ball_at_target = IncomingBall(self.ball_device, target)
        if self.ball_device.config['confirm_eject_type'] == "switch":
            incoming_ball_at_target.add_external_confirm_switch(self.ball_device.config['confirm_eject_switch'])
        elif self.ball_device.config['confirm_eject_type'] == "event":
            incoming_ball_at_target.add_external_confirm_event(self.ball_device.config['confirm_eject_event'])

        target.add_incoming_ball(incoming_ball_at_target)
        return incoming_ball_at_target

    async def _handle_confirm(self, eject_request: OutgoingBall, ball_eject_process: EjectTracker,
                              incoming_ball_at_target: IncomingBall, eject_try: int) -> bool:
        # TODO: check double eject (two balls left). can only happen when not jammed
        timeout = eject_request.eject_timeout
        self.info_log("Wait for confirm with timeout %s", timeout)
        confirm_future = incoming_ball_at_target.wait_for_confirm()
        try:
            await Util.first([confirm_future], timeout=timeout, cancel_others=False)
        except asyncio.TimeoutError:
            self.ball_device.set_eject_state("failed_confirm")
            self.info_log("Got timeout (%ss) before confirm from %s", timeout, eject_request.target)
            return await self._handle_late_confirm_or_missing(eject_request, ball_eject_process,
                                                              incoming_ball_at_target, eject_try)
        else:
            if not confirm_future.done():
                raise AssertionError("Future not done")
            if confirm_future.cancelled():
                raise AssertionError("Eject failed but should not")
            # eject successful
            self.info_log("Got eject confirm")
            await self._handle_eject_success(eject_request)
            return True

    # pylint: disable-msg=too-many-arguments
    async def _handle_playfield_timeout_confirm(self, eject_request, ball_return_future, unknown_balls_future,
                                                incoming_ball_at_target) -> bool:
        await asyncio.sleep(0.1)

        if not ball_return_future.done() and not unknown_balls_future.done():
            # if target is playfield mark eject as confirmed
            self.info_log("Confirming eject because target is playfield and ball did not return.")
            incoming_ball_at_target.ball_arrived()
            await self._handle_eject_success(eject_request)
            return True

        return False

    async def _handle_late_confirm_or_missing(self, eject_request: OutgoingBall, ball_eject_process: EjectTracker,
                                              incoming_ball_at_target: IncomingBall, eject_try: int) -> bool:
        ball_return_future = asyncio.ensure_future(ball_eject_process.wait_for_ball_return())
        unknown_balls_future = asyncio.ensure_future(ball_eject_process.wait_for_ball_unknown_ball())
        eject_success_future = incoming_ball_at_target.wait_for_confirm()
        timeout = self.ball_device.config['ball_missing_timeouts'][eject_request.target] / 1000

        # if ball_eject_process.is_jammed():
        #     # ball returned. eject failed
        #     eject_request.already_left = False
        #     incoming_ball_at_target.did_not_arrive()
        #     return False

        # assume that the ball may have skipped the target device by now
        if self.ball_device.config['confirm_eject_type'] == "target":
            incoming_ball_at_target.set_can_skip()

        if not eject_request.target.is_playfield():
            await eject_request.target.ball_count_handler.wait_for_count_is_valid()
            if eject_success_future.done():
                self.info_log("Got eject confirm (after recounting)")
                await self._handle_eject_success(eject_request)
                return True
        else:
            if (await self._handle_playfield_timeout_confirm(
                    eject_request, ball_return_future, unknown_balls_future,
                    incoming_ball_at_target)):
                return True

        try:
            event = await Util.first([ball_return_future, unknown_balls_future, eject_success_future],
                                     timeout=timeout)
        except asyncio.TimeoutError:
            # handle lost ball
            incoming_ball_at_target.did_not_arrive()
            await self._failed_eject(eject_request, eject_try, True)
            await self.ball_device.lost_ejected_ball(target=eject_request.target)
            # ball is lost but the eject is finished -> return true
            return True

        if event == eject_success_future:
            # we eventually got eject success
            await self._handle_eject_success(eject_request)
            return True
        if event == ball_return_future:
            # ball returned. eject failed
            self.info_log("Ball returned. Eject failed.")
            eject_request.already_left = False
            incoming_ball_at_target.did_not_arrive()
            return False
        if event == unknown_balls_future:
            # TODO: this may be an option
            self.info_log("Got unknown balls. Assuming a ball returned.")
            incoming_ball_at_target.did_not_arrive()
            return False
        # throw an error if we got here
        raise AssertionError("Invalid state")

    async def _handle_eject_success(self, eject_request: OutgoingBall):
        self.info_log("Eject successful")

        await self.machine.events.post_async('balldevice_' + self.ball_device.name + '_ball_eject_success',
                                             balls=1,
                                             target=eject_request.target)
        '''event: balldevice_(name)_ball_eject_success
        config_section: ball_devices
        class_label: ball_device

        desc: One or more balls has successfully ejected from the device
            (name).
        args:
            balls: The number of balls that have successfully ejected.
            target: The target device that has received (or will be receiving)
                the ejected ball(s).
        '''



"""StepStick or similar stepper driver connected to a digital output."""
import asyncio

from typing import Optional

import logging

from mpf.devices.digital_output import DigitalOutput

from mpf.core.platform import StepperPlatform
from mpf.platforms.interfaces.stepper_platform_interface import StepperPlatformInterface
from mpf.core.utility_functions import Util


class DigitalOutputStepStickStepper(StepperPlatformInterface):

    """Stepper on a digital output driven by a StepStick."""

    # pylint: disable-msg=too-many-arguments
    def __init__(self, platform, direction_output, step_output, enable_output, number, config):
        """Initialize stepper."""
        self.platform = platform
        self.number = number
        self._move_complete = asyncio.Event()
        self._move_complete.set()
        self._move_task = None
        self.config = config
        self.direction_output = direction_output    # type: DigitalOutput
        self.step_output = step_output              # type: DigitalOutput
        self.enable_output = enable_output          # type: Optional[DigitalOutput]

    def start_stepper(self, **kwargs):
        """Enable output on init_phase_5 after all hardware has been initialized."""
        del kwargs
        if self.enable_output:
            self.enable_output.enable()

    def move_rel_pos(self, position):
        """Move a number of steps in one direction."""
        if self._move_task and not self._move_task.done():
            raise AssertionError("Last move has not been completed. Calls stop first.")
        self._move_complete.clear()
        self._move_task = self.platform.machine.clock.loop.create_task(self._move_pos(position))
        self._move_task.add_done_callback(Util.raise_exceptions)

    async def _move_pos(self, steps):
        if steps > 0:
            self.direction_output.enable()
        else:
            self.direction_output.disable()
        for _ in range(int(abs(steps))):
            self.step_output.enable()
            await asyncio.sleep(self.config['high_time'])
            self.step_output.disable()
            await asyncio.sleep(self.config['low_time'])

        self._move_complete.set()

    def move_vel_mode(self, velocity):
        """Move at a certain speed."""
        if velocity == 0:
            self.stop()
        else:
            if self._move_task and not self._move_task.done():
                self._move_task.cancel()
            self._move_complete.clear()
            self._move_task = self.platform.machine.clock.loop.create_task(self._move_rel(velocity))
            self._move_task.add_done_callback(Util.raise_exceptions)

    async def _move_rel(self, velocity):
        if velocity > 0:
            self.direction_output.enable()
        else:
            self.direction_output.disable()
        while True:
            self.step_output.enable()
            await asyncio.sleep(self.config['high_time'] * abs(velocity))
            self.step_output.disable()
            await asyncio.sleep(self.config['low_time'] * abs(velocity))
        # this will never complete. you need to call stop

    def stop(self):
        """Stop movements."""
        if self._move_task:
            self._move_task.cancel()
            self._move_task = None
        self.step_output.disable()
        self._move_complete.set()

    def home(self, direction):
        """Not implemented."""
        self.platform.raise_config_error("Please use homing_mode switch", 5, context=self.number)

    async def wait_for_move_completed(self):
        """Wait for move complete."""
        return await self._move_complete.wait()


class StepStickDigitalOutputPlatform(StepperPlatform):

    """Drive a stepper using a StepStick controller on a digital output."""

    def __init__(self, machine):
        """Initialize platform."""
        super().__init__(machine)
        self.log = logging.getLogger('StepStick')

    @classmethod
    def get_stepper_config_section(cls):
        """Return config section."""
        return "step_stick_stepper_settings"

    async def configure_stepper(self, number: str, config: dict) -> "StepperPlatformInterface":
        """Configure a stepper driven by StepStick on a digital output."""
        try:
            direction_output_str, step_output_str, enable_output_str = \
                number.split(":")   # type: str, str, Optional[str]
        except IndexError:
            enable_output_str = None
            try:
                direction_output_str, step_output_str = number.split(":")
            except IndexError:
                return self.raise_config_error("Number for step_stick steppers needs to be "
                                               "direction_output:step_output or "
                                               "direction_output:step_output:enable_output but is {}".format(number),
                                               1)

        try:
            direction_output = self.machine.digital_outputs[direction_output_str]
        except IndexError:
            return self.raise_config_error("direction_output {} does not exist".format(direction_output_str), 2)

        try:
            step_output = self.machine.digital_outputs[step_output_str]
        except IndexError:
            return self.raise_config_error("step_output {} does not exist".format(step_output_str), 3)

        if enable_output_str:
            try:
                enable_output = self.machine.digital_outputs[enable_output_str]     # type: Optional[DigitalOutput]
            except IndexError:
                return self.raise_config_error("enable_output {} does not exist".format(enable_output_str), 4)
        else:
            enable_output = None

        stepper = DigitalOutputStepStickStepper(self, direction_output, step_output, enable_output, number, config)
        self.machine.events.add_handler("init_phase_5", stepper.start_stepper)

        return stepper



"""Test case to start and stop games."""
from unittest.mock import MagicMock

from mpf.tests.MpfTestCase import MpfTestCase


class MpfGameTestCase(MpfTestCase):

    """Test case for starting and running games.

    This is based on ``MpfTestCase`` but adds methods and assertions related
    to running games (rather than just testing MPF components or devices).

    """

    def __init__(self, methodName):
        """Patch minimal config needed to start a game into the machine config.

        This method adds a switch called ``s_start`` with a tag called
        ``start``.

        """
        super().__init__(methodName)
        self.machine_config_patches['switches'] = dict()
        self.machine_config_patches['switches']['s_start'] = {"number": "", "tags": "start"}

    def start_two_player_game(self):
        """Start two player game."""
        self.start_game()
        self.add_player()

    def fill_troughs(self):
        """Fill all ball devices tagged with  ``trough`` with balls."""
        for trough in self.machine.ball_devices.items_tagged("trough"):
            for switch in trough.ball_count_handler.counter.config['ball_switches']:
                self.hit_switch_and_run(switch.name, 0)

        self.advance_time_and_run()

    def start_game(self, num_balls_known=None):
        """Start a game.

        This method checks to make sure a game is not running,
        then hits and releases the ``s_start`` switch, and
        finally checks to make sure a game actually started
        properly.

        For example:

        .. code::

            self.start_game()

        """
        if num_balls_known is not None:
            self.assertNumBallsKnown(num_balls_known)

        # game start should work
        self.assertGameIsNotRunning()
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.assertGameIsRunning()
        self.assertEqual(1, self.machine.game.num_players)
        self.assertPlayerNumber(1)

    def add_player(self):
        """Add a player to the current game.

        This method hits and releases a switch called ``s_start``
        and then verifies that the player count actually increased
        by 1.

        You can call this method multiple times to add multiple
        players. For example, to start a game and then add 2 additional
        players (for 3 players total), you would use:

        .. code::

            self.start_game()
            self.add_player()
            self.add_player()

        """
        prev_players = self.machine.game.num_players
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run(1)
        self.assertEqual(prev_players + 1, self.machine.game.num_players)

    def assertBallNumber(self, number):
        """Asserts that the current ball is a certain ball numebr.

        Args:
            number: The number to check.

        Raises:
            Assertion error if there is no game in progress or if
            the current ball is not the ball number passed.

        The following code will check to make sure the game is on
        Ball 1:

        .. code::

            self.assertBallNumber(1)

        """
        self.assertGameIsRunning()
        self.assertEqual(number, self.machine.game.player.ball)

    def assertBallsInPlay(self, balls):
        """Asserts that a certain number of balls are in play.

        Note that the number of balls in play is not necessarily the same as
        the number of balls on the playfield. (For example, a ball could
        be held in a ball device, or the machine could be in the process
        of adding a ball to the platfield.)

        Args:
            balls: The number of balls you want to assert are in
                play.

        To assert that there are 3 balls in play (perhaps during a multiball),
        you would use:

        .. code::

            self.assertBallsInPlay(3)

        """
        self.assertEqual(balls, self.machine.game.balls_in_play)

    def drain_all_balls(self):
        """Drain all balls in play."""
        drain = self.machine.ball_devices.items_tagged("drain")[0]
        for _ in range(self.machine.game.balls_in_play):
            self.machine.default_platform.add_ball_to_device(drain)

    def drain_one_ball(self):
        """Drain a single ball.

        If more than 1 ball is in play, this method will need to
        be called once for each ball in order to end the current
        ball.

        If you want to drain all balls use `drain_all_balls`.

        """
        drain = self.machine.ball_devices.items_tagged("drain")[0]
        self.machine.default_platform.add_ball_to_device(drain)

    def assertPlayerNumber(self, number):
        """Asserts that the current player is a certain player number.

        Args:
            number: The player number you can to assert is the current
                player.

        For example, to assert that the current player is Player 2, you
        would use:

        .. code::

            self.assertPlayerNumber(2)

        """
        self.assertEqual(number, self.machine.game.player.index + 1)

    def assertPlayerCount(self, count):
        """Asserts that count players exist.

        Args:
            count: The expected number of players.

        For example, to assert that the to players are in the game:

        .. code::

            self.assertPlayerCount(2)

        """
        self.assertEqual(count, len(self.machine.game.player_list))

    def stop_game(self, stop_time=1):
        """Stop the current game.

        This method asserts that a game is running, then call's
        the game mode's ``end_game()`` method, then asserts that
        the game has successfully stopped.

        Example:

        .. code::

            self.stop_game()

        """
        self.assertGameIsRunning()
        self.machine.game.end_game()
        self.advance_time_and_run(stop_time)
        self.assertGameIsNotRunning()

    def assertGameIsRunning(self):
        """Assert a game is running.

        Example:

        .. code::

            self.assertGameIsRunning()

        """
        self.assertIsNotNone(self.machine.game, "Expected a running game but no game is active.")

    def assertGameIsNotRunning(self):
        """Assert a game is not running.

        Example:

        .. code::

            self.assertGameIsNotRunning()

        """
        self.assertIsNone(self.machine.game, "Expected game to have ended but game is active.")



"""Migrate YAML configuration files for MPF from one version to another."""

import argparse
import logging
import os
from datetime import datetime
import errno

from mpf.migrator.migrator import Migrator


class Command:

    """Run the migrator."""

    def __init__(self, mpf_path, machine_path, args):
        """Run the migrator."""
        parser = argparse.ArgumentParser(
            description='Migrates config and show files to the latest version')

        parser.add_argument("-v",
                            action="store_const", dest="consoleloglevel",
                            const=logging.DEBUG,
                            default=logging.INFO,
                            help="Enables verbose logging to the log file")

        parser.add_argument("-l",
                            action="store", dest="logfile",
                            metavar='file_name',
                            default=os.path.join(
                                "logs",
                                datetime.now().strftime(
                                    "migration-%Y-%m-%d-%H-%M-%S.log")),
                            help="The name (and path) of the log file")

        args = parser.parse_args(args)

        try:
            os.makedirs(os.path.join(machine_path, 'logs'))
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise

        logging.basicConfig(level=logging.DEBUG,
                            format='%(name)s : %(message)s',
                            filename=os.path.join(machine_path, args.logfile),
                            filemode='w')

        # sys.stderr
        console = logging.StreamHandler()
        console.setLevel(args.consoleloglevel)

        # set a format which is simpler for console use
        formatter = logging.Formatter('%(levelname)s : %(name)s : %(message)s')

        # tell the handler to use this format
        console.setFormatter(formatter)

        # add the handler to the root logger
        logging.getLogger('').addHandler(console)

        Migrator(mpf_path, machine_path)



from typing import Tuple
from unittest.mock import MagicMock

from mpf.core.bcp.bcp_socket_client import decode_command_string, encode_command_string
from mpf.tests.MpfTestCase import MpfTestCase
from mpf.tests.loop import MockServer, MockQueueSocket


class TestBcp(MpfTestCase):

    def __init__(self, methodName):
        super().__init__(methodName)
        # remove config patch which disables bcp
        del self.machine_config_patches['bcp']
        self.machine_config_patches['bcp'] = dict()
        self.machine_config_patches['bcp']['connections'] = []

    def get_use_bcp(self):
        return True

    def _mock_loop(self):
        self.mock_server = MockServer(self.clock.loop)
        self.clock.mock_server("127.0.0.1", 5051, self.mock_server)

    def _get_and_decode(self, client) -> Tuple[str, dict]:
        data = client.send_queue.get_nowait()
        return decode_command_string(data[0:-1].decode())

    def _encode_and_send(self, client, cmd, **kwargs):
        client.recv_queue.append((encode_command_string(cmd, **kwargs) + '\n').encode())

    def testConnect(self):
        # check that server was opened
        self.assertTrue(self.mock_server.is_bound)

        # add client
        client = MockQueueSocket(self.loop)
        self.machine.clock.loop.run_until_complete(self.mock_server.add_client(client))
        self.advance_time_and_run()

        # check hello
        self.assertEqual(1, client.send_queue.qsize())
        cmd, kwargs = self._get_and_decode(client)
        self.assertEqual("hello", cmd)

        # test trigger
        self.mock_event("test_event")
        client.recv_queue.append((encode_command_string("trigger", name="test_event") + '\n').encode())
        self.advance_time_and_run()
        self.assertEqual(1, self._events['test_event'])

        # register for event/trigger
        client.recv_queue.append((encode_command_string("register_trigger", event="test_trigger") + '\n').encode())
        self.advance_time_and_run()
        self.assertEqual(0, client.send_queue.qsize())

        # post trigger event
        self.post_event("test_trigger")
        self.advance_time_and_run()
        self.assertEqual(1, client.send_queue.qsize())
        cmd, kwargs = self._get_and_decode(client)
        self.assertEqual("trigger", cmd)
        self.assertEqual("test_trigger", kwargs['name'])

        # send goodbye. machine should continue to run.
        client.close = MagicMock()
        client.recv_queue.append((encode_command_string("goodbye") + '\n').encode())
        self.advance_time_and_run()

        client.close.assert_called_with()

        self.assertFalse(self.machine._done)


from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock


class TestBallDevicesEnableCoil(MpfTestCase):
    def get_config_file(self):
        return 'test_enable_coil.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def test_enable_coil_eject(self):
        self.hit_switch_and_run("s_ball1", 0)
        self.hit_switch_and_run("s_ball2", 5)
        self.assertEqual("idle", self.machine.ball_devices["test"].state)

        self.machine.playfield.add_ball()
        self.advance_time_and_run(.1)

        self.assertEqual("enabled", self.machine.coils["eject_coil"].hw_driver.state)
        self.advance_time_and_run(.4)
        self.assertEqual("disabled", self.machine.coils["eject_coil"].hw_driver.state)

        self.release_switch_and_run("s_ball2", 1)

        self.assertEqual("ball_left", self.machine.ball_devices["test"].state)
        self.advance_time_and_run(10)

        self.assertEqual("idle", self.machine.ball_devices["test"].state)

class TestBallDevicesEnableCoilMultiple(MpfTestCase):
    def get_config_file(self):
        return 'test_enable_coil_multiple.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def test_enable_coil_eject_times(self):
        self.hit_switch_and_run("s_ball1", 0)
        self.hit_switch_and_run("s_ball2", 5)
        self.assertEqual("idle", self.machine.ball_devices["test"].state)

        self.machine.playfield.add_ball()
        self.advance_time_and_run(.1)

        self.assertEqual("enabled", self.machine.coils["eject_coil"].hw_driver.state)
        self.advance_time_and_run(.2)
        self.assertEqual("disabled", self.machine.coils["eject_coil"].hw_driver.state)

        self.release_switch_and_run("s_ball2", 1)

        self.assertEqual("ball_left", self.machine.ball_devices["test"].state)
        self.advance_time_and_run(10)

        self.assertEqual("idle", self.machine.ball_devices["test"].state)

        self.machine.playfield.add_ball()
        self.advance_time_and_run(.1)
        self.assertEqual("enabled", self.machine.coils["eject_coil"].hw_driver.state)
        self.advance_time_and_run(.4)
        self.assertEqual("enabled", self.machine.coils["eject_coil"].hw_driver.state)
        self.advance_time_and_run(.2)
        self.assertEqual("disabled", self.machine.coils["eject_coil"].hw_driver.state)

        self.release_switch_and_run("s_ball1", 1)
        self.assertEqual("ball_left", self.machine.ball_devices["test"].state)
        self.advance_time_and_run(10)

        self.assertEqual("idle", self.machine.ball_devices["test"].state)



"""Contains the BallHold device class."""
from collections import deque

from mpf.core.enable_disable_mixin import EnableDisableMixin

from mpf.core.device_monitor import DeviceMonitor
from mpf.core.events import event_handler
from mpf.core.mode_device import ModeDevice
from mpf.core.system_wide_device import SystemWideDevice


@DeviceMonitor("balls_held")
class BallHold(EnableDisableMixin, SystemWideDevice, ModeDevice):

    """Ball hold device which can be used to keep balls in ball devices and control their eject later on."""

    config_section = 'ball_holds'
    collection = 'ball_holds'
    class_label = 'ball_hold'

    __slots__ = ["hold_devices", "source_playfield", "balls_held", "_release_hold", "_released_balls",
                 "hold_queue"]

    def __init__(self, machine, name):
        """Initialise ball hold."""
        self.hold_devices = None
        self.source_playfield = None
        super().__init__(machine, name)

        # initialise variables
        self.balls_held = 0
        self._released_balls = 0
        self._release_hold = None
        self.hold_queue = deque()
        self.machine.events.add_handler('init_phase_3', self._initialize_late)

    @property
    def can_exist_outside_of_game(self):
        """Return true if this device can exist outside of a game."""
        return True

    @classmethod
    def prepare_config(cls, config, is_mode_config):
        """Add default events when outside mode."""
        if not is_mode_config:
            if 'enable_events' not in config:
                config['enable_events'] = 'ball_started'
            if 'disable_events' not in config:
                config['disable_events'] = 'ball_will_end'
        return super().prepare_config(config, is_mode_config)

    async def _initialize(self):
        await super()._initialize()
        self.hold_devices = []
        for device in self.config['hold_devices']:
            self.hold_devices.append(device)

        self.source_playfield = self.config['source_playfield']

    def _initialize_late(self, **kwargs):
        del kwargs
        # postpone until ball devices are ready
        if not self.config['balls_to_hold']:
            self.config['balls_to_hold'] = 0

            for device in self.config['hold_devices']:
                self.config['balls_to_hold'] += device.capacity

    def _enable(self):
        """Enable the hold.

        If the hold is not enabled, no balls will be held.
        """
        self.debug_log("Enabling...")
        self._register_handlers()

    def _disable(self):
        """Disable the hold.

        If the hold is not enabled, no balls will be held.
        """
        self.debug_log("Disabling...")
        self._unregister_handlers()

    @event_handler(1)
    def event_reset(self, **kwargs):
        """Event handler for reset event."""
        del kwargs
        self.reset()

    def reset(self):
        """Reset the hold.

        Will release held balls. Device status will stay the same
        (enabled/disabled). It will wait for those balls to drain and block
        ball_ending until they do. Those balls are not included in ball_in_play.
        """
        self._released_balls += self.release_all()
        self.balls_held = 0

        if self._released_balls > 0:
            # add handler for ball_drain until self._released_balls are drained
            self.machine.events.add_handler(event='ball_drain',
                                            handler=self._wait_for_drain)

            # block ball_ending
            self.machine.events.add_handler(event='ball_ending', priority=10000,
                                            handler=self._block_during_drain)

    def _wait_for_drain(self, balls, **kwargs):
        del kwargs
        if balls <= 0:
            return {'balls': balls}

        if balls > self._released_balls:
            ball_to_reduce = self._released_balls
        else:
            ball_to_reduce = balls

        self._released_balls -= ball_to_reduce
        self.debug_log("%s ball of hold drained.", ball_to_reduce)

        if self._released_balls <= 0:
            if self._release_hold:
                self._release_hold.clear()
                self._release_hold = None
            self.debug_log("All released balls of ball_hold drained.")
            self.machine.events.remove_handler_by_event('ball_ending', self._wait_for_drain)
            self.machine.events.remove_handler_by_event('ball_drain', self._block_during_drain)

        return {'balls': balls - ball_to_reduce}

    def _block_during_drain(self, queue, **kwargs):
        del kwargs
        if self._released_balls > 0:
            queue.wait()
            self._release_hold = queue

    @event_handler(9)
    def event_release_one_if_full(self, **kwargs):
        """Event handler for release_one_if_full event."""
        del kwargs
        self.release_one_if_full()

    def release_one_if_full(self):
        """Release one ball if hold is full."""
        if self.is_full():
            self.release_one()

    @event_handler(8)
    def event_release_one(self, **kwargs):
        """Event handler for release_one event."""
        del kwargs
        self.release_one()

    def release_one(self):
        """Release one ball."""
        self.release_balls(balls_to_release=1)

    @event_handler(7)
    def event_release_all(self, **kwargs):
        """Event handler for release_all event."""
        del kwargs
        self.release_all()

    def release_all(self):
        """Release all balls in hold."""
        return self.release_balls(self.balls_held)

    def release_balls(self, balls_to_release):
        """Release all balls and return the actual amount of balls released.

        Args:
        ----
            balls_to_release: number of ball to release from hold
        """
        if not self.hold_queue:
            return 0

        remaining_balls_to_release = balls_to_release

        self.debug_log("Releasing up to %s balls from hold", balls_to_release)
        balls_released = 0
        while self.hold_queue:
            device, balls_held = self.hold_queue.pop()
            balls = balls_held
            balls_in_device = device.balls
            if balls > balls_in_device:
                balls = balls_in_device

            if balls > remaining_balls_to_release:
                self.hold_queue.append(
                    (device, balls_held - remaining_balls_to_release))
                balls = remaining_balls_to_release

            device.eject(balls=balls)
            balls_released += balls
            remaining_balls_to_release -= balls
            if remaining_balls_to_release <= 0:
                break

        if balls_released > 0:
            self.machine.events.post(
                'ball_hold_' + self.name + '_balls_released',
                balls_released=balls_released)
            '''event: ball_hold_(name)_balls_released

            desc: The ball hold device (name) has just released a ball(s).

            args:
                balls_released: The number of balls that were just released.
            '''

        self.balls_held -= balls_released
        return balls_released

    def _register_handlers(self):
        priority = (self.mode.priority if self.mode else 0) + \
            self.config['priority']
        # register on ball_enter of hold_devices
        for device in self.hold_devices:
            self.machine.events.add_handler(
                'balldevice_' + device.name + '_ball_enter',
                self._hold_ball, device=device, priority=priority)

    def _unregister_handlers(self):
        # unregister ball_enter handlers
        self.machine.events.remove_handler(self._hold_ball)

    def is_full(self):
        """Return true if hold is full."""
        return self.remaining_space_in_hold() == 0

    def remaining_space_in_hold(self):
        """Return the remaining capacity of the hold."""
        balls = self.config['balls_to_hold'] - self.balls_held
        if balls < 0:
            balls = 0
        return balls

    def _hold_ball(self, device, new_balls, unclaimed_balls, **kwargs):
        """Handle result of _ball_enter event of hold_devices."""
        del new_balls
        del kwargs
        # if full do not take any balls
        if self.is_full():
            self.debug_log("Cannot hold balls. Hold is full.")
            return {'unclaimed_balls': unclaimed_balls}

        # if there are no balls do not claim anything
        if unclaimed_balls <= 0:
            return {'unclaimed_balls': unclaimed_balls}

        capacity = self.remaining_space_in_hold()
        # take ball up to capacity limit
        if unclaimed_balls > capacity:
            balls_to_hold = capacity
        else:
            balls_to_hold = unclaimed_balls

        self.balls_held += balls_to_hold
        self.debug_log("Held %s balls", balls_to_hold)

        # post event for ball capture
        self.machine.events.post('ball_hold_' + self.name + '_held_ball',
                                 balls_held=balls_to_hold,
                                 total_balls_held=self.balls_held)
        '''event: ball_hold_(name)_held_ball
        desc: The ball hold device (name) has just held additional ball(s).

        args:
            balls_held: The number of new balls just held.
            total_balls_held: The current total number of balls this device
                has held.
        '''

        # check if we are full now and post event if yes
        if self.is_full():
            self.machine.events.post('ball_hold_' + self.name + '_full',
                                     balls=self.balls_held)
        '''event: ball_hold_(name)_full
        desc: The ball hold device (name) is now full.
        args:
            balls: The number of balls currently held in this device.
        '''

        self.hold_queue.append((device, unclaimed_balls))

        return {'unclaimed_balls': unclaimed_balls - balls_to_hold}



"""Generic list randomizer."""
from uuid import uuid4
import random


class Randomizer:

    """Generic list randomizer."""

    def __init__(self, items, machine=None, template_type=None):
        """Initialise Randomizer."""
        self.fallback_value = None
        self.force_different = True
        self.force_all = False
        self.disable_random = False
        # self.loop - property which sets force_all=True if loop==False
        self.items = list()

        self._loop = True
        self._template_type = template_type
        self.data = None
        self._uuid = uuid4()

        if isinstance(items, (list, tuple)):
            for i in items:
                if isinstance(i, (tuple, list)):
                    this_item = i[0]
                    this_weight = int(i[1])
                else:
                    this_item = i
                    this_weight = 1
                # If a template_type is provided, convert the event into a conditional template
                if machine and template_type:
                    this_item = self.generate_template(machine, template_type, this_item)
                self.items.append((this_item, this_weight))

        elif isinstance(items, dict):
            for this_item, this_weight in items.items():
                # If a template_type is provided, convert the event into a conditional template
                if machine and template_type:
                    this_item = self.generate_template(machine, template_type, this_item)
                self.items.append((this_item, int(this_weight)))
                self.items.sort(key=lambda x: x[0].name or x[0])
        else:
            raise AssertionError("Invalid input for Randomizer")

        self.data = dict()
        self._init_data(self.data)

    def __iter__(self):
        """Return iterator."""
        return self

    def __next__(self, conditional_args=None):
        """Return next."""
        if not conditional_args:
            conditional_args = {}

        if self.disable_random:
            return self._next_not_random(conditional_args)

        potential_nexts = list()
        items = self._get_items(conditional_args)

        if self.force_all:
            potential_nexts = [
                x for x in items if x[0] not in self.data['items_sent']]

        elif self.force_different:
            potential_nexts = [
                x for x in items if x[0] is not self.data['current_item']]

        if not potential_nexts:

            if not self._loop:
                raise StopIteration

            self.data['items_sent'] = set()

            # force different only works with more than 1 elements
            if self.force_different and len(self.items) > 1:
                potential_nexts = [x for x in items if x[0] is not (
                    self.data['current_item'])]
            else:
                potential_nexts = list(items)

        # If no values were found due to all conditions failing, return the fallback
        if not potential_nexts:
            return self.fallback_value

        self.data['current_item'] = self.pick_weighted_random(potential_nexts)
        self.data['items_sent'].add(self.data['current_item'])

        return self.data['current_item']

    @property
    def loop(self):
        """Return loop property."""
        return self._loop

    @loop.setter
    def loop(self, loop):
        """Set loop property."""
        if loop:
            self._loop = True
        else:
            self._loop = False
            self.force_all = True

    def _next_not_random(self, conditional_args):
        if self.data['current_item_index'] == len(self.items):
            if not self.loop:
                raise StopIteration

            self.data['current_item_index'] = 0

        self.data['current_item'] = (
            self._get_items(conditional_args)[self.data['current_item_index']][0])

        self.data['current_item_index'] += 1

        return self.data['current_item']

    @staticmethod
    def _init_data(data_dict):
        """Initialise dict."""
        data_dict['current_item'] = None
        data_dict['items_sent'] = set()
        data_dict['current_item_index'] = 0  # only used with disable random

    def get_current(self):
        """Return current item."""
        if self.data['current_item']:
            return self.data['current_item']

        return self.__next__()

    def get_next(self, conditional_args=None):
        """Return next item."""
        return self.__next__(conditional_args)

    def _get_items(self, conditional_args):
        if self._template_type:
            conditional_items = list()
            for event, weight in self.items:
                if not event.condition:
                    conditional_items.append((event.name or event, weight))
                elif event.condition.evaluate(conditional_args):
                    conditional_items.append((event.name, weight))
            return conditional_items
        return self.items

    @staticmethod
    def generate_template(machine, template_type, value):
        """Convert a string with conditions into a conditional event template object."""
        if template_type == "event":
            return machine.placeholder_manager.parse_conditional_template(value)
        # Add additional template_type support here, as needed
        return value

    @staticmethod
    def pick_weighted_random(items):
        """Pick a random item.

        Args:
        ----
            items: Items to select from
        """
        total_weights = sum([x[1] for x in items])
        value = random.randint(1, total_weights)
        index_value = 0

        for item in items:
            index_value += item[1]
            if index_value >= value:
                return item[0]

        return items[-1][0]



"""Interface for switches."""
import abc
from typing import Any
from mpf.core.platform import SwitchConfig


class SwitchPlatformInterface(metaclass=abc.ABCMeta):

    """Interface for switches in hardware platforms.

    SwitchPlatformInterface is an abstract base class that should be overridden for all
    switches interface classes on supported platforms.  This class ensures the proper required
    methods are implemented to support switch operations in MPF.
    """

    __slots__ = ["config", "number"]

    def __init__(self, config: SwitchConfig, number: Any) -> None:
        """Initialise default attributes for switches."""
        self.config = config    # type: SwitchConfig
        self.number = number    # type: Any

    @abc.abstractmethod
    def get_board_name(self):
        """Return the name of the board of this driver."""
        raise NotImplementedError

    def __repr__(self):
        """Return board + number."""
        return "<Switch {} {} (config: {})>".format(self.get_board_name(), self.number, self.config)



"""Config file error in a MPF config file."""
from mpf.exceptions.base_error import BaseError


class ConfigFileError(BaseError):

    """Error in a config file found."""

    def get_short_name(self):
        """Return short name."""
        return "CFE"

    def get_long_name(self):
        """Return long name."""
        return "Config File Error"



from mpf.tests.MpfGameTestCase import MpfGameTestCase
from mpf.tests.MpfTestCase import test_config


class TestBallDeviceSmartVirtual(MpfGameTestCase):
    def get_config_file(self):
        return 'test_ball_device.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def get_platform(self):
        return 'smart_virtual'

    def test_eject(self):
        # add initial balls to trough
        self.hit_switch_and_run("s_ball_switch1", 1)
        self.hit_switch_and_run("s_ball_switch2", 1)
        self.assertEqual(2, self.machine.ball_devices["test_trough"].balls)
        self.assertEqual(2, self.machine.ball_devices["test_trough"].available_balls)

        # call eject
        self.machine.ball_devices["test_trough"].eject()
        self.assertEqual(2, self.machine.ball_devices["test_trough"].balls)
        self.assertEqual(1, self.machine.ball_devices["test_trough"].available_balls)

        # one ball should be gone
        self.advance_time_and_run(30)
        self.assertEqual(1, self.machine.ball_devices["test_trough"].balls)
        self.assertEqual(1, self.machine.ball_devices["test_trough"].available_balls)

    def test_eject_all(self):
        # add initial balls to trough
        self.hit_switch_and_run("s_ball_switch1", 1)
        self.hit_switch_and_run("s_ball_switch2", 1)
        self.assertEqual(2, self.machine.ball_devices["test_trough"].balls)
        self.assertEqual(2, self.machine.ball_devices["test_trough"].available_balls)

        # call eject_all
        self.machine.ball_devices["test_trough"].eject_all()
        self.advance_time_and_run(30)

        # all balls should be gone
        self.assertEqual(0, self.machine.ball_devices["test_trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["test_trough"].available_balls)

    @test_config("test_player_controlled_eject.yaml")
    def test_player_controlled_eject(self):
        self.start_game()
        self.advance_time_and_run(10)
        self.assertBallsInPlay(1)
        self.assertBallsOnPlayfield(0)

        self.hit_and_release_switch("s_launch_button")
        self.advance_time_and_run(5)
        self.assertBallsOnPlayfield(1)



import inspect

from mpf.core.machine import MachineController
from mpf.tests.MpfTestCase import MpfTestCase


class MockConfigPlayers(MpfTestCase):

    """Add all config players without installing mpf-mc."""

    @staticmethod
    def _load_mc_players(cls):
        mc_players = {
            "sound_player": "mpfmc.config_players.plugins.sound_player",
            "widget_player": "mpfmc.config_players.plugins.widget_player",
            "slide_player": "mpfmc.config_players.plugins.slide_player",
            "display_lights": "mpfmc.config_players.plugins.display_light_player",
            "track_player": "mpfmc.config_players.plugins.track_player",
            "sound_loop_player": "mpfmc.config_players.plugins.sound_loop_player",
        }

        for name, module in mc_players.items():
            imported_module = inspect.importlib.import_module(module)
            setattr(cls, '{}_player'.format(name),
                    imported_module.player_cls(cls))

    def setUp(self):
        MachineController._register_plugin_config_players = self._load_mc_players
        super().setUp()


class BaseMpfMachineTestCase(MockConfigPlayers):

    def get_enable_plugins(self):
        return True

    def get_config_file(self):
        return "config.yaml"

    def get_machine_path(self):
        return ""

    def get_absolute_machine_path(self):
        # do not use path relative to MPF folder
        return self.get_machine_path()

    def get_platform(self):
        return 'smart_virtual'


class MpfMachineTestCase(BaseMpfMachineTestCase):

    """MPF only machine test case."""

    def get_use_bcp(self):
        return True

    def __init__(self, methodName='runTest'):
        super().__init__(methodName)

        # remove config patches
        self.machine_config_patches = dict()
        # use bcp mock
        self.machine_config_patches['bcp'] = \
            {"connections": {"local_display": {"type": "mpf.tests.MpfBcpTestCase.MockBcpClient"}}, "servers": []}

        # increase test expected duration
        self.expected_duration = 5.0



"""MPF CLI commands."""
import argparse
from importlib import import_module
import os
import sys
from pkg_resources import iter_entry_points


import mpf.core
from mpf._version import version

EXAMPLES_FOLDER = 'examples'
CONFIG_FOLDER = 'config'


class MpfCommandLineParser:

    """Base class for cli commands."""

    def __init__(self, args, path):
        """Initialise CLI entry point."""
        self.argv = args
        self.path = path
        self.mpf_path = os.path.abspath(os.path.join(mpf.core.__path__[0],
                                                     os.pardir))

    def get_machine_path(self, machine_path_hint=None) -> str:
        """Find the full machine path based on the current directory and option hint.

        Args:
        ----
            machine_path_hint: Helps MPF locate the machine path. If None,
                the 'config' folder in the current working directory is used.

        Returns a string of full path of the machine folder that was located.
        """
        machine_path = None

        if machine_path_hint:
            if os.path.isdir(os.path.join(self.path, machine_path_hint)):
                # If the path hint resolves to a folder, use that as the
                # machine folder
                machine_path = os.path.join(self.path, machine_path_hint)

            else:
                # If the folder is invalid, see if we have an examples machine
                # folder with that name
                example_machine_path = os.path.abspath(os.path.join(
                    self.mpf_path, os.pardir, EXAMPLES_FOLDER,
                    machine_path_hint))

                if os.path.isdir(example_machine_path):
                    machine_path = example_machine_path

        else:
            # no path hint passed.
            # Is there a /config folder in our current folder? If so we assume
            # the current folder is the machine folder
            if os.path.isdir(os.path.join(self.path, CONFIG_FOLDER)):
                machine_path = self.path

        if machine_path:
            return machine_path

        if machine_path_hint:
            wrong_path = os.path.abspath(machine_path_hint)
        else:
            wrong_path = os.path.abspath(os.curdir)

        raise AssertionError("Error: Could not find machine in folder: '{}'. "
                             "Either start MPF from within your machine root folder or provide the path after the "
                             "command.".format(wrong_path))

    def parse_args(self):
        """Parse command line arguments."""
        parser = argparse.ArgumentParser(description='MPF Command')

        parser.add_argument("machine_path", help="Path of the machine folder.",
                            default=None, nargs='?')

        parser.add_argument("--version",
                            action="version", version=version,
                            help="Displays the MPF, config file, and BCP "
                                 "version info and exits")

        # the problem with parser.add_argument is it will take the first
        # positional argument it finds for machine_path and set it to the
        # machine path, regardless of what's in front of it. So for example,
        # args of "-c step4" will lead to machine_path='step4', but that's not
        # right, machine_path should be None. But this is because it doesn't
        # know that -c wants to consume the next positional arg.

        # So our workaround is we check if there are any argv, and if so, we
        # check to see if the first one starts with a dash, meaning it's an
        # optional arg and guaranteeing that whatever's after it is NOT our
        # machine path, so in that case, we just insert a None as the machine
        # path in front of it and everything is cool.

        if len(self.argv) > 1 and self.argv[1].startswith('-'):
            self.argv.insert(1, None)

        args, remaining_args = parser.parse_known_args(self.argv[1:])
        machine_path = self.get_machine_path(args.machine_path)

        return machine_path, remaining_args


class CommandLineUtility(MpfCommandLineParser):

    """Default CLI entry point."""

    def __init__(self, path=None):
        """Initialise CLI entry point."""
        super().__init__(path=path, args=sys.argv[:])
        self.external_commands = dict()
        self.get_external_commands()

    def get_external_commands(self):
        """Entry point to hook more commands.

        This is used from mpf mc.
        """
        for entry_point in iter_entry_points(group='mpf.command', name=None):
            command, function_ref = entry_point.load()()
            self.external_commands[command] = function_ref

    @classmethod
    def check_python_version(cls):
        """Check that we have at least Python 3."""
        if sys.version_info[0] != 3:
            print("MPF requires Python 3. You have Python {}.{}.{}".format(
                sys.version_info[0], sys.version_info[1], sys.version_info[2]
            ))
            sys.exit()

    def execute(self):
        """Execute the command that was just set up."""
        self.check_python_version()

        commands = set()

        for file in os.listdir(os.path.join(self.mpf_path, 'commands')):
            commands.add(os.path.splitext(file)[0])

        command = 'game'

        if len(self.argv) > 1:

            if self.argv[1] in self.external_commands:
                command = self.argv.pop(1)
                self.external_commands[command](self.mpf_path,
                                                *self.parse_args())
                return
            if self.argv[1] in commands:
                command = self.argv.pop(1)

        _module = import_module('mpf.commands.%s' % command)

        if hasattr(_module, "SUBCOMMAND") and _module.SUBCOMMAND:
            _module.Command(self.argv, self.path)
        else:
            machine_path, remaining_args = self.parse_args()

            _module.Command(self.mpf_path, machine_path, remaining_args)


def run_from_command_line(args=None):
    """Run a CLI command.

    Args:
    ----
        args: Command line arguments that were passed.

    """
    del args
    path = os.path.abspath(os.path.curdir)
    CommandLineUtility(path).execute()



from mpf.core.mode import Mode


class Mode3(Mode):
    def mode_init(self):
        self.custom_code = True

    def mode_start(self, **kwargs):
        pass

    def mode_stop(self, **kwargs):
        pass



"""Score Queue player for SS style scoring."""
import re
from typing import Dict, List, Any

from mpf.core.config_player import ConfigPlayer


class ScoreQueuePlayer(ConfigPlayer):

    """SS style scoring based on config."""

    config_file_section = 'score_queue_player'
    show_section = 'score_queues'

    __slots__ = []  # type: List[str]

    @staticmethod
    def is_entry_valid_outside_mode(settings) -> bool:
        """Score queue is only valid in game."""
        del settings
        return False

    def play(self, settings: dict, context: str, calling_context: str,
             priority: int = 0, **kwargs) -> None:
        """Variable name."""
        for var, s in settings.items():
            if s['condition'] and not s['condition'].evaluate(kwargs):
                continue
            self.machine.score_queues[var].score(s['int'].evaluate(kwargs))

    def validate_config_entry(self, settings: dict, name: str) -> dict:
        """Validate one entry of this player."""
        config = {}
        if not isinstance(settings, dict):
            self.raise_config_error("Settings of score_queue_player should "
                                    "be a dict. But are: {}".format(settings), 5, context=name)
        for var, s in settings.items():
            var_conditional_event = self.machine.placeholder_manager.parse_conditional_template(var)
            value_dict = self._parse_config(s, name)
            value_dict["condition"] = var_conditional_event.condition
            config[var_conditional_event.name] = value_dict
            if not bool(re.match('^[0-9a-zA-Z_-]+$', var_conditional_event.name)):
                self.raise_config_error("Variable may only contain letters, numbers, dashes and underscores. "
                                        "Name: {}".format(var_conditional_event.name), 4, context=name)
        return config

    def get_express_config(self, value: Any) -> dict:
        """Parse express config."""
        return {"int": value}



"""Switch controller which handles all switches in MPF.

Contains the SwitchController class which is responsible for reading switch
states and posting events to the framework.
"""
from collections import namedtuple
import asyncio
from functools import partial
from typing import Any, Callable, Dict, List, Tuple

from mpf.core.platform import SwitchPlatform

from mpf.core.machine import MachineController
from mpf.core.mpf_controller import MpfController
from mpf.devices.switch import Switch

MonitoredSwitchChange = namedtuple("MonitoredSwitchChange", ["name", "label", "platform", "num", "state"])
SwitchHandler = namedtuple("SwitchHandler", ["switch_name", "callback", "state", "ms"])
TimedSwitchHandler = namedtuple("TimedSwitchHandler", ["callback", 'state', 'ms'])


class RegisteredSwitch:

    """Registered switch handler."""

    __slots__ = ["ms", "callback", "cancelled"]

    def __init__(self, ms, callback):
        """Initialise registered switch."""
        self.ms = ms
        self.callback = callback
        self.cancelled = False


class SwitchController(MpfController):

    """Tracks all switches in the machine, receives switch activity, and converts switch changes into events."""

    config_name = "switch_controller"

    __slots__ = ["registered_switches", "_timed_switch_handler_delay", "_active_timed_switches",
                 "_switch_lookup", "monitors", "_initialised"]

    def __init__(self, machine: MachineController) -> None:
        """Initialise switch controller."""
        super().__init__(machine)
        self.registered_switches = dict()                       # type: Dict[Switch, List[List[RegisteredSwitch]]]
        # Dictionary of switches and states that have been registered for
        # callbacks.

        self._timed_switch_handler_delay = dict()               # type: Any

        self._active_timed_switches = {}                        # type: Dict[str, Dict[float, List[TimedSwitchHandler]]]
        # Dictionary of switches that are currently in a state counting ms
        # waiting to notify their handlers. In other words, this is the dict
        # that tracks current switches for things like "do foo() if switch bar
        # is active for 100ms."

        self._switch_lookup = dict()                            # type: Dict[Tuple[str, SwitchPlatform], Switch]
        # Lookup table for switch + platform to an Switch object

        # register for events
        self.machine.events.add_async_handler('init_phase_2', self._initialize_switches, 1000)
        # priority 1000 so this fires first

        self.monitors = list()      # type: List[Callable[[MonitoredSwitchChange], None]]

        # to detect early switch changes before init
        self._initialised = False

    def register_switch(self, switch: Switch):
        """Add a switch object to the switch controller for tracking.

        Args:
        ----
            switch: Switch object to add
        """
        self.registered_switches[switch] = [list(), list()]

    async def _initialize_switches(self, **kwargs):
        del kwargs
        await self.update_switches_from_hw()

        for switch in self.machine.switches.values():
            # build lookup table
            self._switch_lookup[(switch.hw_switch.number, switch.platform)] = switch

        self._initialised = True

        self.log_active_switches()

    async def update_switches_from_hw(self):
        """Update the states of all the switches be re-reading the states from the hardware platform.

        This method works silently and does not post any events if any switches
        changed state.
        """
        # create a list of hw switch numbers, platforms, and switch objects
        platforms = set()
        switches = set()  # (switch_object, number)

        for switch in self.machine.switches.values():
            platforms.add(switch.platform)
            switches.add((switch, switch.hw_switch.number))

        for platform in platforms:
            switch_states = await platform.get_hw_switch_states()

            for switch, number in switches:
                # if two platforms have the same number choose the right switch
                if switch.platform != platform:
                    continue
                try:
                    switch.state = switch_states[number] ^ switch.invert
                except (IndexError, KeyError):
                    raise AssertionError("Missing switch {} in update from HW: {} "
                                         "State from HW: {}, switches known: {}".
                                         format(switch, platform, switch_states, switches))

    async def verify_switches(self) -> bool:
        """Verify that switches states match the hardware.

        Loops through all the switches and queries their hardware states via
        their platform interfaces and then compares that to the state that MPF
        thinks the switches are in.

        Throws logging warnings if anything doesn't match.

        This method is notification only. It doesn't fix anything.
        """
        current_states = dict()

        for switch in self.machine.switches.values():
            current_states[switch] = switch.state

        await self.update_switches_from_hw()

        ok = True

        for switch in self.machine.switches.values():
            if switch.state != current_states[switch]:  # pragma: no cover
                ok = False
                self.warning_log("Switch State Error! Switch: %s, HW State: "
                                 "%s, MPF State: %s", switch.name,
                                 current_states[switch],
                                 switch.state)

        return ok

    def is_state(self, switch: Switch, state, ms=0.0):
        """Check if switch is in state.

        Query whether a switch is in a given state and (optionally)
        whether it has been in that state for the specified number of ms.

        Args:
        ----
            switch: Switch object to check.
            state: Bool of the state to check. True is active and False is
                inactive.
            ms: Milliseconds that the switch has been in that state. If this
                is non-zero, then this method will only return True if the
                switch has been in that state for at least the number of ms
                specified.

        Returns: True if the switch_name has been in the state for the given
            number of ms. If ms is not specified, returns True if the switch
            is in the state regardless of how long it's been in that state.
        """
        if not self._initialised:
            raise AssertionError("Cannot read switch state before init_phase_3")
        if not ms:
            ms = 0.0

        if ms:
            return switch.state == state and ms <= switch.get_ms_since_last_change()

        return switch.state == state

    def is_active(self, switch, ms=None):
        """Query whether a switch is active.

        Args:
        ----
            switch: Switch object to check.
            ms: Milliseconds that the switch has been active. If this
                is non-zero, then this method will only return True if the
                switch has been in that state for at least the number of ms
                specified.

        Returns: True if the switch_name has been active for the given
            number of ms. If ms is not specified, returns True if the switch
            is in the state regardless of how long it's been in that state.
        """
        if not self._initialised:
            raise AssertionError("Cannot read switch state before init_phase_3")
        if not ms:
            ms = 0.0

        if ms:
            return switch.state == 1 and ms <= switch.get_ms_since_last_change()

        return switch.state == 1

    def is_inactive(self, switch, ms=None):
        """Query whether a switch is inactive.

        Args:
        ----
            switch: Switch object to check.
            ms: Milliseconds that the switch has been inactive. If this
                is non-zero, then this method will only return True if the
                switch has been in that state for at least the number of ms
                specified.

        Returns: True if the switch_name has been inactive for the given
            number of ms. If ms is not specified, returns True if the switch
            is in the state regardless of how long it's been in that state.
        """
        if not self._initialised:
            raise AssertionError("Cannot read switch state before init_phase_3")
        if not ms:
            ms = 0.0

        if ms:
            return switch.state == 0 and ms <= switch.get_ms_since_last_change()

        return switch.state == 0

    # pylint: disable-msg=too-many-arguments
    def process_switch_by_num(self, num, state, platform, logical=False, timestamp=None):
        """Process a switch state change by switch number.

        Args:
        ----
            num: The switch number (based on the platform number) for the
                switch you're setting.
            state: The state to set, either 0 or 1.
            platform: The platform this switch is on.
            logical: Whether the state you're setting is the logical or
                physical state of the switch. If a switch is NO (normally
                open), then the logical and physical states will be the same.
                NC (normally closed) switches will have physical and
                logical states that are inverted from each other.
            timestamp: Timestamp when this switch change happened.

        """
        if not self._initialised:
            raise AssertionError("Got early switch change for switch {} to state {}. platform: {}".format(
                num, state, platform))
        switch = self._switch_lookup.get((num, platform), None)

        if switch:
            self.process_switch_obj(switch, state, logical, timestamp)
        else:
            if self._debug_to_console or self._debug_to_file:
                self.debug_log("Unknown switch %s change to state %s on platform %s", num, state, platform)
            # if the switch is not configured still trigger the monitor
            for monitor in self.monitors:
                monitor(MonitoredSwitchChange(name=str(num), label="{}-{}".format(str(platform), str(num)),
                                              platform=platform, num=str(num), state=state))

    def process_switch(self, name, state, logical=False, timestamp=None):
        """Process a new switch state change for a switch by name.

        This is the method that is called by the platform driver whenever a
        switch changes state. It's also used by the "other" modules that
        activate switches, including the keyboard and OSC interfaces.

        State 0 means the switch changed from active to inactive, and 1 means
        it changed from inactive to active. (The hardware & platform code
        handles NC versus NO switches and translates them to 'active' versus
        'inactive'.)

        Args:
        ----
            name: The string name of the switch.
            state: Boolean or int of state of the switch you're processing,
                True/1 is active, False/0 is inactive.
            logical: Boolean which specifies whether the 'state' argument
                represents the "physical" or "logical" state of the switch. If
                True, a 1 means this switch is active and a 0 means it's
                inactive, regardless of the NC/NO configuration of the switch.
                If False, then the state parameter passed will be inverted if
                the switch is configured to be an 'NC' type. Typically the
                hardware will send switch states in their raw (logical=False)
                states, but other interfaces like the keyboard and OSC will use
                logical=True.
            timestamp: Timestamp when this switch change happened.

        """
        if self._debug:
            self.debug_log("Processing switch. Name: %s, state: %s, logical: %s,", name, state, logical)

        if timestamp is None:
            timestamp = self.machine.clock.get_time()

        try:
            obj = self.machine.switches[name]
        except KeyError:    # pragma: no cover
            raise AssertionError('Cannot process switch "{}" as this is not a valid switch name.'.format(name))

        self.process_switch_obj(obj, state, logical, timestamp)

    def process_switch_obj(self, obj: Switch, state, logical, timestamp=None):
        """Process a new switch state change for a switch by name.

        Args:
        ----
            obj: The switch object.
            state: Boolean or int of state of the switch you're processing,
                True/1 is active, False/0 is inactive.
            logical: Boolean which specifies whether the 'state' argument
                represents the "physical" or "logical" state of the switch. If
                True, a 1 means this switch is active and a 0 means it's
                inactive, regardless of the NC/NO configuration of the switch.
                If False, then the state parameter passed will be inverted if
                the switch is configured to be an 'NC' type. Typically the
                hardware will send switch states in their raw (logical=False)
                states, but other interfaces like the keyboard and OSC will use
                logical=True.
            timestamp: Timestamp when this switch change happened.

        This is the method that is called by the platform driver whenever a
        switch changes state. It's also used by the "other" modules that
        activate switches, including the keyboard and OSC interfaces.

        State 0 means the switch changed from active to inactive, and 1 means
        it changed from inactive to active. (The hardware & platform code
        handles NC versus NO switches and translates them to 'active' versus
        'inactive'.)
        """
        assert obj.hw_switch is not None
        # We need int, but this lets it come in as boolean also
        if state:
            state = 1
        else:
            state = 0

        if timestamp is None:
            timestamp = self.machine.clock.get_time()

        # flip the logical & physical states for NC switches
        hw_state = state

        if obj.invert:
            if logical:  # NC + logical means hw_state is opposite of state
                hw_state ^= 1
            else:
                # NC w/o logical (i.e. hardware state was sent) means logical
                # state is the opposite
                state ^= 1

        # if the switch is already in this state, then abort
        if obj.state == state:
            if not self.machine.options['production']:
                self.warning_log(
                    "Received duplicate switch state %s for switch %s from the platform interface.",
                    state, obj.name, error_no=1)
            return

        # Update the hardware state since we always want this to match real hw
        obj.hw_state = hw_state
        # update the switch device
        obj.state = state
        obj.last_change = timestamp

        if state:
            self.info_log("<<<<<<< '%s' active >>>>>>>", obj.name)
        else:
            self.info_log("<<<<<<< '%s' inactive >>>>>>>", obj.name)

        self._cancel_timed_handlers(obj)

        self._call_handlers(obj, state)

        for monitor in self.monitors:
            monitor(MonitoredSwitchChange(name=obj.name, label=obj.label, platform=obj.platform,
                                          num=obj.hw_switch.number, state=state))

    def wait_for_switch(self, switch: Switch, state: int = 1, only_on_change=True, ms=0):
        """Wait for a switch to change into a state.

        Args:
        ----
            switch: String to wait for.
            state: The state to wait for. 0 = inactive, 1 = active, 2 = opposite to current.
            only_on_change: Bool which controls whether this wait will be
                triggered now if the switch is already in the state, or
                whether it will wait until the switch changes into that state.
            ms: How long the switch needs to be in the new state to trigger
                the wait.

        """
        return self.wait_for_any_switch([switch], state, only_on_change, ms)

    def wait_for_any_switch(self, switches: List[Switch], state: int = 1, only_on_change=True, ms=0):
        """Wait for the first switch in the list to change into state.

        Args:
        ----
            switches: Iterable of switches. Whichever switch changes first will trigger this wait.
            state: The state to wait for. 0 = inactive, 1 = active, 2 = opposite to current.
            only_on_change: Bool which controls whether this wait will be
                triggered now if the switch is already in the state, or
                whether it will wait until the switch changes into that state.
            ms: How long the switch needs to be in the new state to trigger
                the wait.

        """
        future = asyncio.Future()   # type: asyncio.Future

        if not only_on_change:
            for switch in switches:
                if self.is_state(switch, state, ms):
                    future.set_result({"switch_name": switch.name, "state": state, "ms": ms})
                    return future

        handlers = []   # type: List[SwitchHandler]
        future.add_done_callback(partial(self._future_done, handlers))      # type: ignore
        for switch in switches:
            if state == 2:
                handler_state = 0 if self.is_active(switch) else 1
            else:
                handler_state = state
            handlers.append(self.add_switch_handler_obj(switch, state=handler_state, ms=ms,
                                                        callback=partial(self._wait_handler,
                                                                         ms=ms,
                                                                         _future=future,
                                                                         switch_name=switch.name)))
        return future

    def _future_done(self, handlers: List[SwitchHandler], future: asyncio.Future):
        del future
        for handler in handlers:
            self.remove_switch_handler_by_key(handler)

    @staticmethod
    def _wait_handler(_future: asyncio.Future, **kwargs):
        if not _future.done():
            _future.set_result(kwargs)

    def _cancel_timed_handlers(self, switch):
        """Cancel all timed handlers as they became invalid since the switch state changed."""
        if switch not in self._active_timed_switches:
            return

        # just nuke the list of handlers
        del self._active_timed_switches[switch]
        # unregister handler
        if switch in self._timed_switch_handler_delay:
            self.machine.clock.unschedule(self._timed_switch_handler_delay[switch][0])
            del self._timed_switch_handler_delay[switch]

    def _add_timed_switch_handler(self, switch, time: float, timed_switch_handler: TimedSwitchHandler):
        if switch not in self._active_timed_switches:
            self._active_timed_switches[switch] = {time: [timed_switch_handler]}
            next_event_time = time
        elif time not in self._active_timed_switches[switch]:
            self._active_timed_switches[switch][time] = [timed_switch_handler]
            next_event_time = min(self._active_timed_switches[switch].keys())
        else:
            self._active_timed_switches[switch][time].append(timed_switch_handler)
            next_event_time = min(self._active_timed_switches[switch].keys())

        add_handler = False
        if switch not in self._timed_switch_handler_delay:
            add_handler = True
        elif next_event_time < self._timed_switch_handler_delay[switch][1]:
            add_handler = True
            self.machine.clock.unschedule(self._timed_switch_handler_delay[switch][0])

        if add_handler:
            handler = self.machine.clock.loop.call_at(
                next_event_time, partial(self._process_active_timed_switches, switch))
            self._timed_switch_handler_delay[switch] = (handler, next_event_time)

    def _call_handlers(self, switch, state):
        for entry in self.registered_switches[switch][state][:]:  # generator?
            # Found an entry.

            # skip if the handler has been removed in the meantime
            if entry.cancelled:
                continue

            if entry.ms:
                # This entry is for a timed switch, so add it to our
                # active timed switch list
                key = switch.last_change + (entry.ms / 1000.0)
                value = TimedSwitchHandler(callback=entry.callback,
                                           state=state,
                                           ms=entry.ms)
                self._add_timed_switch_handler(switch, key, value)
                if self._debug_to_console or self._debug_to_file:
                    self.debug_log(
                        "Found timed switch handler for k/v %s / %s",
                        key, value)
            else:
                # This entry doesn't have a timed delay, so do the action
                # now
                entry.callback()

    def add_monitor(self, monitor: Callable[[MonitoredSwitchChange], None]):
        """Add a monitor callback which is called on switch changes."""
        if monitor not in self.monitors:
            self.monitors.append(monitor)

    def remove_monitor(self, monitor: Callable[[MonitoredSwitchChange], None]):
        """Remove a monitor callback."""
        if monitor in self.monitors:
            self.monitors.remove(monitor)

    # pylint: disable-msg=too-many-arguments
    def add_switch_handler(self, switch_name, callback, state=1, ms=0,
                           return_info=False, callback_kwargs=None) -> SwitchHandler:
        """Register a handler to take action on a switch event.

        Args:
        ----
            switch_name: String name of the switch you're adding this handler
                for.
            callback: The method you want called when this switch handler fires.
            state: Integer of the state transition you want to callback to be
                triggered on. Default is 1 which means it's called when the
                switch goes from inactive to active, but you can also use 0
                which means your callback will be called when the switch becomes
                inactive
            ms: Integer. If you specify a 'ms' parameter, the handler won't be
                called until the witch is in that state for that many
                milliseconds.
            return_info: If True, the switch controller will pass the
                parameters of the switch handler as arguments to the callback,
                including switch_name, state, and ms. If False (default), it
                just calls the callback with no parameters.
            callback_kwargs: Additional kwargs that will be passed with the
                callback.

        You can mix & match entries for the same switch here.
        """
        switch = self.machine.switches[switch_name]
        return self.add_switch_handler_obj(switch, callback, state, ms, return_info, callback_kwargs)

    def add_switch_handler_obj(self, switch, callback, state=1, ms=0,
                               return_info=False, callback_kwargs=None):
        """Register a handler to take action on a switch event.

        Same as add_switch_handler but you can pass a switch object instead of a name.
        """
        if callback_kwargs and return_info:
            callback = partial(callback, switch_name=switch.name, state=state, ms=ms, **callback_kwargs)
        elif return_info:
            callback = partial(callback, switch_name=switch.name, state=state, ms=ms)
        elif callback_kwargs:
            callback = partial(callback, **callback_kwargs)

        if self._debug_to_console or self._debug_to_file:
            self.debug_log("Registering switch handler: %s, %s, state: %s, ms: %s"
                           ", info: %s", switch.name, callback,
                           state, ms, return_info)

        entry_val = RegisteredSwitch(ms=ms, callback=callback)
        self.registered_switches[switch][state].append(entry_val)

        # If the switch handler that was just registered has a delay (i.e. ms>0,
        # then let's see if the switch is currently in the state that the
        # handler was registered for. If so, and if the switch has been in this
        # state for less time than the ms registered, then we need to add this
        # switch to our active_timed_switches list so this handler is called
        # when this switch's active time expires. (in other words, we're
        # catching delayed switches that were in progress when this handler was
        # registered.
        if ms:  # only do this for handlers that have delays
            current_time = self.machine.clock.get_time()
            if switch.last_change > current_time - ms and state == switch.state:
                # figure out when this handler should fire based on the
                # switch's original activation time.
                key = switch.last_change + (ms / 1000.0)
                value = TimedSwitchHandler(callback=callback,
                                           state=state,
                                           ms=ms)
                self._add_timed_switch_handler(switch, key, value)

        # Return the args we used to setup this handler for easy removal later
        return SwitchHandler(switch, callback, state, ms)

    def remove_switch_handler_by_key(self, switch_handler: SwitchHandler):
        """Remove switch handler by key returned from add_switch_handler."""
        self.remove_switch_handler_obj(switch_handler.switch_name, switch_handler.callback, switch_handler.state,
                                       switch_handler.ms)

    def remove_switch_handler_by_keys(self, switch_handlers: List[SwitchHandler]):
        """Remove switch handlers by list of keys returned from add_switch_handler."""
        for switch_handler in switch_handlers:
            self.remove_switch_handler_obj(switch_handler.switch_name, switch_handler.callback, switch_handler.state,
                                           switch_handler.ms)

    def remove_switch_handler(self, switch_name, callback, state=1, ms=0):
        """Remove a registered switch handler.

        Currently this only works if you specify everything exactly as you set
        it up. (Except for return_info, which doesn't matter if true or false,
        it will remove either / both.
        """
        switch = self.machine.switches[switch_name]
        self.remove_switch_handler_obj(switch, callback, state, ms)

    def remove_switch_handler_obj(self, switch, callback, state=1, ms=0):
        """Remove a registered switch handler.

        Same as remove_switch_handler but takes a switch object instead of the name.
        """
        if self._debug_to_console or self._debug_to_file:
            self.debug_log(
                "Removing switch handler. Switch: %s, State: %s, ms: %s",
                switch.name, state, ms)

        for entry in list(self.registered_switches[switch][state]):
            if entry.ms == ms and entry.callback == callback:
                entry.cancelled = True
                self.registered_switches[switch][state].remove(entry)

        if switch in self._active_timed_switches:
            for k in list(self._active_timed_switches[switch].keys()):
                timed_entry = self._active_timed_switches[switch][k]
                for dummy_key, entry in enumerate(timed_entry):
                    if entry.state == state and entry.ms == ms and entry.callback == callback:
                        del self._active_timed_switches[switch][k][dummy_key]

    def log_active_switches(self, **kwargs):
        """Write out entries to the INFO log file of all switches that are currently active."""
        del kwargs
        for k, v in self._switch_lookup.items():
            if v.state:
                self.info_log("Found active switch: %s %s", k, v)

    @staticmethod
    def get_active_event_for_switch(switch_name):
        """Return the event name which is posted when switch_name becomes active."""
        return "{}_active".format(switch_name)

    def _process_active_timed_switches(self, switch):
        """Process active times switches.

        Checks the current list of active timed switches to see if it's
        time to take action on any of them. If so, does the callback and then
        removes that entry from the list.
        """
        del self._timed_switch_handler_delay[switch]
        next_event_time = False
        current_time = self.machine.clock.get_time()
        for k in list(self._active_timed_switches[switch].keys()):
            if k <= current_time:  # change to generator?
                for entry in list(self._active_timed_switches[switch][k]):
                    # check if removed by previous entry
                    if entry not in self._active_timed_switches[switch][k]:
                        continue
                    if self._debug_to_console or self._debug_to_file:
                        self.debug_log(
                            "Processing timed switch handler. Switch: %s "
                            " State: %s, ms: %s", switch.name,
                            entry.state, entry.ms)
                    entry.callback()
                del self._active_timed_switches[switch][k]
            else:
                if not next_event_time or next_event_time > k:
                    next_event_time = k

        self.machine.events.process_event_queue()
        if next_event_time:
            handler = self.machine.clock.loop.call_at(
                next_event_time,
                partial(self._process_active_timed_switches, switch))
            self._timed_switch_handler_delay[switch] = (handler, next_event_time)



"""Interface defines for LISY."""


class LisyDefines:

    """LISY messages."""

    InfoGetConnectedLisyHardware = 0
    InfoLisyVersion = 1
    InfoGetApiVersion = 2
    InfoGetNumberOfLamps = 3
    InfoGetNumberOfSolenoids = 4
    InfoGetNumberOfSounds = 5
    InfoGetNumberOfDisplays = 6
    InfoGetDisplayDetails = 7
    InfoGetGameInfo = 8
    InfoGetSwitchCount = 9

    LampsGetStatusOfLamps = 10
    LampsSetLampOn = 11
    LampsSetLampOff = 12
    FadeModernLights = 13

    GetModernLightsCount = 19

    SolenoidsGetStatusOfSolenoid = 20
    SolenoidsSetSolenoidToOn = 21
    SolenoidsSetSolenoidToOff = 22
    SolenoidsPulseSolenioid = 23
    SolenoidsSetSolenoidPulseTime = 24
    SetSolenoidsRecycleTime = 25
    PulseAndEnableSolenoidWithPWM = 26

    DisplaysSetDisplay0To = 30
    DisplaysSetDisplay1To = 31
    DisplaysSetDisplay2To = 32
    DisplaysSetDisplay3To = 33
    DisplaysSetDisplay4To = 34
    DisplaysSetDisplay5To = 35
    DisplaysSetDisplay6To = 36

    SwitchesGetStatusOfSwitch = 40
    SwitchesGetChangedSwitches = 41

    SoundPlaySound = 50
    SoundStopAllSounds = 51
    SoundPlaySoundFile = 52
    SoundTextToSpeech = 53
    SoundSetVolume = 54

    ConfigureHardwareRuleForSolenoid = 60

    GeneralReset = 100
    GeneralWatchdog = 101



"""Case insensitive dict."""
# Based on this: http://stackoverflow.com/questions/2082152/case-insensitive-dictionary
from typing import List


class CaseInsensitiveDict(dict):

    """A dict which lowercases all keys."""

    __slots__ = []  # type: List[str]

    @staticmethod
    def lower(key):
        """Lowercase the key."""
        return key.lower() if isinstance(key, str) else key

    def __init__(self, *args, **kwargs):
        """Initialise case insensitve dict."""
        super().__init__(*args, **kwargs)
        self._convert_keys()

    def __getitem__(self, key):
        """Return item for key."""
        return super().__getitem__(self.__class__.lower(key))

    def __setitem__(self, key, value):
        """Set item for key to value."""
        super().__setitem__(self.__class__.lower(key), value)

    def __delitem__(self, key):
        """Delete item for key."""
        return super().__delitem__(self.__class__.lower(key))

    def __contains__(self, key):
        """Check if dict contains a key."""
        return super().__contains__(self.__class__.lower(key))

    def pop(self, key, *args, **kwargs):
        """Retrieve and delete a value for a key."""
        return super().pop(self.__class__.lower(key), *args, **kwargs)

    def get(self, key, *args, **kwargs):
        """Return item for key."""
        return super().get(self.__class__.lower(key), *args, **kwargs)

    def setdefault(self, key, *args, **kwargs):
        """Set defaults."""
        return super().setdefault(self.__class__.lower(key), *args, **kwargs)

    def update(self, e=None, **f):
        """Update a value for a key."""
        super().update(self.__class__(e))
        super().update(self.__class__(**f))

    def _convert_keys(self):
        for k in list(self.keys()):
            v = super().pop(k)
            self.__setitem__(k, v)
